# pylint: disable=unsubscriptable-object
import unittest
from unittest import TestCase
from unittest.mock import Mock, patch

import numpy as np
from AnyQt.QtCore import QMimeData, QPoint, QPointF, Qt
from AnyQt.QtGui import QDragEnterEvent, QDropEvent, QDrag
from AnyQt.QtWidgets import QApplication

from orangewidget.tests.base import GuiTest

from Orange.data import Table, Domain, \
    ContinuousVariable, DiscreteVariable, StringVariable
from Orange.widgets.settings import ContextSetting
from Orange.widgets.utils import vartype
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.data.owselectcolumns \
    import OWSelectAttributes, VariablesListItemModel, \
    SelectAttributesDomainContextHandler, SelectedVarsView, PrimitivesView
from Orange.widgets.data.owrank import OWRank
from Orange.widgets.utils.itemmodels import select_rows
from Orange.widgets.widget import AttributeList

Continuous = vartype(ContinuousVariable("c"))
Discrete = vartype(DiscreteVariable("d"))


# It is, what it is (and should be), pylint: disable=invalid-name
class TestSelectAttributesDomainContextHandler(TestCase):
    def setUp(self):
        self.domain = Domain(
            attributes=[ContinuousVariable('c1'),
                        DiscreteVariable('d1', values='abc'),
                        DiscreteVariable('d2', values='def')],
            class_vars=[DiscreteVariable('d3', values='ghi')],
            metas=[ContinuousVariable('c2'),
                   DiscreteVariable('d4', values='jkl')]
        )
        self.args = (self.domain,
                     {'c1': Continuous, 'd1': Discrete,
                      'd2': Discrete, 'd3': Discrete},
                     {'c2': Continuous, 'd4': Discrete, })

        self.handler = SelectAttributesDomainContextHandler(first_match=False)
        self.handler.read_defaults = lambda: None

    def test_open_context(self):
        # Why not? pylint: disable=use-dict-literal
        self.handler.bind(SimpleWidget)
        context = Mock(
            attributes=self.args[1], metas=self.args[2], values=dict(
                domain_role_hints=({('d1', Discrete): ('available', 0),
                                    ('d2', Discrete): ('meta', 0),
                                    ('c1', Continuous): ('attribute', 0),
                                    ('d3', Discrete): ('attribute', 1),
                                    ('d4', Discrete): ('attribute', 2),
                                    ('c2', Continuous): ('class', 0)}, -2),
                with_metas=[('d1', Discrete), ('d2', Discrete)]
            ))
        self.handler.global_contexts = \
            [Mock(values={}), context, Mock(values={})]

        widget = SimpleWidget()
        self.handler.initialize(widget)
        domain = self.args[0]
        self.handler.open_context(widget, domain)
        self.assertEqual(widget.domain_role_hints,
                         {domain['d1']: ('available', 0),
                          domain['d2']: ('meta', 0),
                          domain['c1']: ('attribute', 0),
                          domain['d3']: ('attribute', 1),
                          domain['d4']: ('attribute', 2),
                          domain['c2']: ('class', 0)})

    def test_open_context_with_imperfect_match(self):
        # Why not? pylint: disable=use-dict-literal
        self.handler.bind(SimpleWidget)
        context1 = Mock(values=dict(
            domain_role_hints=({('d1', Discrete): ('attribute', 0),
                                ('m2', Discrete): ('meta', 0)}, -2)
        ))
        context = Mock(values=dict(
            domain_role_hints=({('d1', Discrete): ('available', 0),
                                ('d2', Discrete): ('meta', 0),
                                ('c1', Continuous): ('attribute', 0),
                                ('d6', Discrete): ('attribute', 1),
                                ('d7', Discrete): ('attribute', 2),
                                ('c2', Continuous): ('class', 0)}, -2)
        ))
        self.handler.global_contexts = \
            [Mock(values={}), context1, context, Mock(values={})]

        widget = SimpleWidget()
        self.handler.initialize(widget)
        domain = self.args[0]
        self.handler.open_context(widget, domain)

        self.assertEqual(widget.domain_role_hints,
                         {domain['d1']: ('available', 0),
                          domain['d2']: ('meta', 0),
                          domain['c1']: ('attribute', 0),
                          domain['c2']: ('class', 0)})


class TestModel(TestCase):
    def setUp(self):
        self.variables = \
            [ContinuousVariable(c) for c in "xyz"] + \
            [StringVariable(s) for s in "spqr"] + \
            [DiscreteVariable(d, values=tuple("def")) for d in "abc"]

    @staticmethod
    def _vars(s):
        return "".join(var.name for var in s)

    def test_drop_mime(self):
        m = VariablesListItemModel(self.variables)
        mime = m.mimeData([m.index(1, 0)])
        self.assertTrue(mime.hasFormat(VariablesListItemModel.MIME_TYPE))
        assert m.dropMimeData(mime, Qt.MoveAction, 5, 0, m.index(-1, -1))
        self.assertIs(m[5], m[1])
        assert m.dropMimeData(mime, Qt.MoveAction, -1, -1, m.index(-1, -1))
        self.assertIs(m[11], m[1])

    def test_drop_mime_primitive(self):
        mime = QMimeData()
        # the encoded 'data' is empty, variables are passed by properties
        mime.setData(VariablesListItemModel.MIME_TYPE, b'')
        mime.setProperty("_items", self.variables[2:])

        m = VariablesListItemModel(self.variables[:2], primitive=False)
        assert m.dropMimeData(mime, Qt.MoveAction, 1, 0, m.index(-1, -1))
        self.assertEqual(self._vars(m), "xzspqrabcy")
        self.assertTrue(mime.property("_moved"))

        m = VariablesListItemModel(self.variables[:2], primitive=True)
        assert m.dropMimeData(mime, Qt.MoveAction, 1, 0, m.index(-1, -1))
        self.assertEqual(self._vars(m), "xzabcy")
        self.assertEqual(self._vars(mime.property("_moved")), "zabc")

    def test_drop_mime_noop(self):
        m = VariablesListItemModel(self.variables[:2], primitive=False)

        mime = QMimeData()
        # the encoded 'data' is empty, variables are passed by properties
        mime.setData(VariablesListItemModel.MIME_TYPE, b'')

        mime.setProperty("_items", self.variables[:2])
        self.assertTrue(m.dropMimeData(mime, Qt.IgnoreAction, 1, 0, m.index(-1, -1)))
        self.assertEqual(self._vars(m), "xy")
        self.assertIsNone(mime.property("_moved"))

        mime.setProperty("_items", None)
        self.assertFalse(m.dropMimeData(mime, Qt.MoveAction, 1, 0, m.index(-1, -1)))
        self.assertEqual(self._vars(m), "xy")
        self.assertIsNone(mime.property("_moved"))

        mime = QMimeData()
        mime.setData("application/x-that-other-format", b'')
        mime.setProperty("_items", self.variables[:2])

        self.assertFalse(m.dropMimeData(mime, Qt.MoveAction, 1, 0, m.index(-1, -1)))
        self.assertEqual(self._vars(m), "xy")
        self.assertIsNone(mime.property("_moved"))

    def test_mimedata(self):
        m = VariablesListItemModel(self.variables)
        mime = m.mimeData([m.index(i, 0) for i in (1, 2, 5, 7, 9)])
        # 0123456789
        # xyzspqrabc
        self.assertEqual(self._vars(mime.property("_items")), "yzqac")

    def test_flags(self):
        m = VariablesListItemModel([ContinuousVariable("X")])
        flags = m.flags(m.index(0))
        self.assertTrue(flags & Qt.ItemIsDragEnabled)
        self.assertFalse(flags & Qt.ItemIsDropEnabled)
        # 'invalid' index is drop enabled -> indicates insertion capability
        flags = m.flags(m.index(-1, -1))
        self.assertTrue(flags & Qt.ItemIsDropEnabled)


class TestViews(GuiTest):
    def setUp(self):
        self.variables = \
            [ContinuousVariable(c) for c in "xyz"] + \
            [StringVariable(s) for s in "spqr"] + \
            [DiscreteVariable(d, values=tuple("def")) for d in "abc"]
        self.model = VariablesListItemModel(self.variables)
        self.view = SelectedVarsView()
        self.view.setModel(self.model)

    @staticmethod
    def _vars(s):
        return "".join(var.name for var in s)

    @patch("AnyQt.QtGui.QDrag.exec")
    def test_noop(self, drag_exec):
        with patch.object(self.view, "selectedIndexes", return_value=[]):
            assert self.view.startDrag(Qt.MoveAction) is None
            drag_exec.assert_not_called()

        with patch.object(self.view, "selectedIndexes",
                          return_value=[self.model.index(1, 0)]), \
                patch.object(self.model, "mimeData", return_value=None):
            assert self.view.startDrag(Qt.MoveAction) is None
            drag_exec.assert_not_called()

    def test_move(self):

        def drag_exec(self, *_):
            self.mimeData().setProperty("_moved", moved)
            return Qt.MoveAction

        # 0123456789
        # xyzspqrabc
        #  yz p rab
        indexes = [self.model.index(i, 0) for i in (1, 2, 4, 6, 7, 8)]
        selmodel = self.view.selectionModel()
        for index in indexes:
            selmodel.select(index, selmodel.Select)
        with patch("AnyQt.QtGui.QDrag.exec", drag_exec):

            moved = None
            self.view.startDrag(Qt.MoveAction)
            self.assertEqual(self.model.rowCount(), 10)

            moved = True
            self.view.startDrag(Qt.MoveAction)
            self.assertEqual(self._vars(self.model), "xsqc")

            self.model[:] = self.variables
            indexes = [self.model.index(i, 0) for i in (1, 2, 4, 6, 7, 8)]
            for index in indexes:
                selmodel.select(index, selmodel.Select)
            moved = [self.model[i] for i in (4, 6)]
            self.view.startDrag(Qt.MoveAction)
            self.assertEqual(self._vars(self.model), "xyzsqabc")

    @patch("AnyQt.QtGui.QDropEvent.source")
    def test_primitives_accepts_drop(self, src):
        view = PrimitivesView()
        mime = QMimeData()
        mime.setData(VariablesListItemModel.MIME_TYPE, b'')
        event = QDropEvent(QPointF(20, 20), Qt.MoveAction, mime,
                          Qt.NoButton, Qt.NoModifier)

        with patch.object(event, "mimeData"):
            self.assertFalse(view.acceptsDropEvent(event))
            event.mimeData.assert_not_called()
            self.assertFalse(event.isAccepted())

        src.return_value.window.return_value = view.window()

        mime.setProperty("_items", self.variables)
        self.assertTrue(view.acceptsDropEvent(event))
        self.assertTrue(event.isAccepted())
        event.setAccepted(False)

        mime.setProperty("_items", None)
        self.assertFalse(view.acceptsDropEvent(event))
        self.assertFalse(event.isAccepted())

        mime.setProperty("_items", [])
        self.assertFalse(view.acceptsDropEvent(event))
        self.assertFalse(event.isAccepted())

        mime.setProperty("_items", self.variables[3:7])  # string variables
        self.assertFalse(view.acceptsDropEvent(event))
        self.assertFalse(event.isAccepted())


class SimpleWidget:
    domain_role_hints = ContextSetting({})
    required = ContextSetting("", required=ContextSetting.REQUIRED)

    def retrieveSpecificSettings(self):
        pass

    def storeSpecificSettings(self):
        pass


class TestOWSelectAttributes(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWSelectAttributes)

    def assertVariableCountsEqual(self, available, used, classattrs, metas=0):
        self.widget.update_interface_state()
        for (name, box, view), nattrs in zip(self.widget.view_boxes,
                                              (available, used, classattrs, metas)):
            self.assertEqual(view.model().rowCount(), nattrs)
            if nattrs:
                self.assertEqual(box.title(), f"{name} ({nattrs})")
            else:
                self.assertEqual(box.title(), name)

    def assertControlsEnabled(self, _list, button, box, widget=None):
        if widget is None:
            widget = self.widget
        control = widget.use_features_box
        self.assertEqual(control.button.isEnabled(), button)
        self.assertEqual(control.isVisibleTo(widget), box)
        self.assertEqual(widget.used_attrs_view.isEnabled(), _list)
        self.assertEqual(widget.move_attr_button.isEnabled(), _list)
        if button:
            control.button.click()
            self.assertEqual(control.button.isEnabled(), False)

    def test_multiple_target_variable(self):
        """
        More than one target variable can be moved to a box for target variables
        at the same time and moved back as well.
        GH-2100
        GH-2086
        """
        iris = Table("iris")
        self.send_signal(self.widget.Inputs.data, iris)
        self.assertVariableCountsEqual(available=0, used=4, classattrs=1)

        self.widget.move_class_button.click()
        self.assertVariableCountsEqual(available=0, used=4, classattrs=1)

        self.widget.used_attrs_view.selectAll()
        self.widget.move_selected(self.widget.used_attrs_view)
        self.assertVariableCountsEqual(available=4, used=0, classattrs=1)

        self.widget.available_attrs_view.selectAll()
        self.widget.move_selected(self.widget.class_attrs_view)
        self.assertVariableCountsEqual(0, 0, 5)

    def test_move_to_primitive(self):
        app = QApplication.instance()
        widget = self.widget

        data = Table("zoo")
        self.send_signal(widget.Inputs.data, data)

        # Selecting meta attribute must enable the corresponding button
        widget.meta_attrs_view.selectAll()
        app.processEvents()
        self.assertFalse(widget.move_attr_button.isEnabled())
        self.assertFalse(widget.move_class_button.isEnabled())
        self.assertTrue(widget.move_meta_button.isEnabled())

        # Moving to available
        widget.move_meta_button.click()
        self.assertVariableCountsEqual(available=1, used=16, classattrs=1, metas=0)

        # Selecting available attributes must enable only meta button
        # because all selected attrs are non-primitive and can't be used for
        # features or classes
        widget.available_attrs_view.selectAll()
        app.processEvents()
        self.assertFalse(widget.move_attr_button.isEnabled())
        self.assertFalse(widget.move_class_button.isEnabled())
        self.assertTrue(widget.move_meta_button.isEnabled())

        # Selecting class attributes must enable the corresponding button
        widget.class_attrs_view.selectAll()
        app.processEvents()
        self.assertFalse(widget.move_attr_button.isEnabled())
        self.assertTrue(widget.move_class_button.isEnabled())
        self.assertFalse(widget.move_meta_button.isEnabled())

        # Move it to available
        widget.move_class_button.click()
        self.assertVariableCountsEqual(available=2, used=16, classattrs=0, metas=0)

        # Selecting meta attributes: nothing there, so disable all buttons
        widget.meta_attrs_view.selectAll()
        app.processEvents()
        self.assertFalse(widget.move_attr_button.isEnabled())
        self.assertFalse(widget.move_class_button.isEnabled())
        self.assertFalse(widget.move_meta_button.isEnabled())

        # Selecting available attributes must now enable all buttons because
        # there some of selected attributes are not primitive
        widget.available_attrs_view.selectAll()
        app.processEvents()
        self.assertTrue(widget.move_attr_button.isEnabled())
        self.assertTrue(widget.move_class_button.isEnabled())
        self.assertTrue(widget.move_meta_button.isEnabled())

        # Move to metas should move both attributes
        widget.move_meta_button.click()
        self.assertVariableCountsEqual(available=0, used=16, classattrs=0, metas=2)

        # Move them back to available
        widget.meta_attrs_view.selectAll()
        app.processEvents()
        widget.move_meta_button.click()
        self.assertVariableCountsEqual(available=2, used=16, classattrs=0, metas=0)

        # Now move them to class: only one should be moved
        widget.available_attrs_view.selectAll()
        app.processEvents()
        widget.move_class_button.click()
        self.assertVariableCountsEqual(available=1, used=16, classattrs=1, metas=0)

        # Move them back to available
        widget.class_attrs_view.selectAll()
        app.processEvents()
        widget.move_class_button.click()
        self.assertVariableCountsEqual(available=2, used=16, classattrs=0, metas=0)

        # Now move them to attributes: only one should be moved
        widget.available_attrs_view.selectAll()
        app.processEvents()
        widget.move_attr_button.click()
        self.assertVariableCountsEqual(available=1, used=17, classattrs=0, metas=0)

    def test_input_features(self):
        data = Table("zoo")
        in_features = AttributeList(data.domain.attributes)
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.features, in_features)
        self.assertControlsEnabled(True, False, True)
        self.assertVariableCountsEqual(0, 16, 1, 1)
        self.assertListEqual(self.get_output(self.widget.Outputs.features),
                             list(data.domain.attributes))

    def test_input_features_by_name(self):
        data = Table("zoo")
        in_features = AttributeList([DiscreteVariable(attr.name, attr.values)
                                     for attr in data.domain.attributes])
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.features, in_features)
        self.assertControlsEnabled(True, False, True)
        self.assertVariableCountsEqual(0, 16, 1, 1)
        self.assertListEqual(self.get_output(self.widget.Outputs.features),
                             list(data.domain.attributes))

    def test_input_features_same_domain(self):
        data = Table("zoo")
        in_features = AttributeList(data.domain.variables + data.domain.metas)
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.features, in_features)
        self.assertControlsEnabled(True, False, True)
        self.assertVariableCountsEqual(0, 16, 1, 1)
        self.assertListEqual(self.get_output(self.widget.Outputs.features),
                             list(data.domain.attributes))

    def test_input_features_sub_domain(self):
        data = Table("zoo")
        in_features = AttributeList(data.domain.attributes[::3])
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.features, in_features)
        self.assertControlsEnabled(True, True, True)
        self.assertVariableCountsEqual(10, 6, 1, 1)
        self.assertListEqual(self.get_output(self.widget.Outputs.features),
                             in_features)

    def test_input_features_by_name_sub_domain(self):
        data = Table("zoo")
        in_features = AttributeList([DiscreteVariable(attr.name, attr.values)
                                     for attr in data.domain.attributes[:5]])
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.features, in_features)
        self.assertControlsEnabled(True, True, True)
        self.assertVariableCountsEqual(11, 5, 1, 1)
        self.assertListEqual(self.get_output(self.widget.Outputs.features),
                             list(data.domain.attributes[:5]))

    def test_input_features_diff_domain(self):
        zoo = Table("zoo")
        in_features = AttributeList(Table("iris").domain.attributes)
        self.send_signal(self.widget.Inputs.data, zoo)
        self.send_signal(self.widget.Inputs.features, in_features)
        self.assertControlsEnabled(True, False, False)
        self.assertVariableCountsEqual(0, 16, 1, 1)
        self.assertListEqual(self.get_output(self.widget.Outputs.features),
                             list(zoo.domain.attributes))
        self.assertTrue(self.widget.Warning.mismatching_domain.is_shown())
        self.send_signal(self.widget.Inputs.features, None)
        self.assertFalse(self.widget.Warning.mismatching_domain.is_shown())

    def test_input_features_no_data(self):
        data = Table("zoo")
        in_features = AttributeList(data.domain.variables + data.domain.metas)
        self.send_signal(self.widget.Inputs.features, in_features)
        self.assertControlsEnabled(True, False, False)
        self.assertVariableCountsEqual(0, 0, 0, 0)
        self.assertIsNone(self.get_output(self.widget.Outputs.features))

    def test_input_combinations(self):
        data = Table("iris")
        in_features = AttributeList(data.domain.attributes[:2])

        # check initial state
        self.assertControlsEnabled(True, False, False)
        self.assertVariableCountsEqual(0, 0, 0, 0)
        self.assertIsNone(self.get_output(self.widget.Outputs.features))

        # send data
        self.send_signal(self.widget.Inputs.data, data)
        self.assertControlsEnabled(True, False, False)
        self.assertVariableCountsEqual(0, 4, 1, 0)
        self.assertEqual(len(self.get_output(self.widget.Outputs.features)), 4)

        # send features
        self.send_signal(self.widget.Inputs.features, in_features)
        self.assertControlsEnabled(True, True, True)
        self.assertVariableCountsEqual(2, 2, 1, 0)
        self.assertEqual(len(self.get_output(self.widget.Outputs.features)), 2)

        # remove data
        self.send_signal(self.widget.Inputs.data, None)
        self.assertControlsEnabled(True, False, False)
        self.assertVariableCountsEqual(0, 0, 0, 0)
        self.assertIsNone(self.get_output(self.widget.Outputs.features))

        # remove features
        self.send_signal(self.widget.Inputs.features, None)
        self.assertControlsEnabled(True, False, False)
        self.assertVariableCountsEqual(0, 0, 0, 0)
        self.assertIsNone(self.get_output(self.widget.Outputs.features))

        # send features
        self.send_signal(self.widget.Inputs.features, in_features)
        self.assertControlsEnabled(True, False, False)
        self.assertVariableCountsEqual(0, 0, 0, 0)
        self.assertIsNone(self.get_output(self.widget.Outputs.features))

        # send data
        self.send_signal(self.widget.Inputs.data, data)
        self.assertControlsEnabled(True, False, True)
        self.assertVariableCountsEqual(2, 2, 1, 0)
        self.assertEqual(len(self.get_output(self.widget.Outputs.features)), 2)

        # remove features
        self.send_signal(self.widget.Inputs.features, None)
        self.assertControlsEnabled(True, False, False)
        self.assertVariableCountsEqual(2, 2, 1, 0)
        self.assertEqual(len(self.get_output(self.widget.Outputs.features)), 2)

    def test_input_features_from_rank(self):
        data = Table("iris")
        owrank = self.create_widget(OWRank)
        self.send_signal(owrank.Inputs.data, data, widget=owrank)
        rank_features = self.get_output(owrank.Outputs.features, widget=owrank)
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.features, rank_features)
        self.assertControlsEnabled(True, True, True)
        features = self.get_output(self.widget.Outputs.features)
        self.assertListEqual(rank_features, features)

    def test_use_features_checked(self):
        data = Table("iris")
        attrs = data.domain.attributes

        # prepare stored settings (check "Use input features")
        in_features = AttributeList(attrs[:2])
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.features, in_features)
        self.widget.use_features_box.checkbox.setChecked(True)
        self.assertControlsEnabled(False, False, True)
        out_features = self.get_output(self.widget.Outputs.features)
        self.assertListEqual(out_features, in_features)
        settings = self.widget.settingsHandler.pack_data(self.widget)

        # "Use input features" is checked by default
        widget = self.create_widget(OWSelectAttributes, settings)
        in_features = AttributeList(attrs[:3])
        self.send_signal(widget.Inputs.data, data, widget=widget)
        self.send_signal(widget.Inputs.features, in_features, widget=widget)
        self.assertControlsEnabled(False, False, True, widget)
        out_features = self.get_output(widget.Outputs.features, widget=widget)
        self.assertListEqual(out_features, in_features)

        # reset "Features"
        widget.reset()
        out_features = self.get_output(widget.Outputs.features, widget=widget)
        self.assertFalse(widget.use_features_box.checkbox.isChecked())
        self.assertListEqual(out_features, AttributeList(attrs))
        self.assertControlsEnabled(True, True, True, widget)

    def test_used_attrs_supported_types(self):
        data = Table("zoo")
        event = self._drag_enter_event(data.domain[:1])
        self.widget.used_attrs_view.dragEnterEvent(event)
        self.assertTrue(event.isAccepted())

        event = self._drag_enter_event(data.domain.metas)
        self.widget.used_attrs_view.dragEnterEvent(event)
        self.assertFalse(event.isAccepted())

    def _drag_enter_event(self, variables):
        # pylint: disable=attribute-defined-outside-init
        self.event_data = mime = QMimeData()
        mime.setProperty("_items", variables)
        return QDragEnterEvent(QPoint(0, 0), Qt.MoveAction, mime,
                               Qt.NoButton, Qt.NoModifier)

    def test_move_rows(self):
        data = Table("iris")[:5]
        w = self.widget
        self.send_signal(w.Inputs.data, data)
        view = w.used_attrs_view
        model = view.model()
        selmodel = view.selectionModel()
        midx = model.index(1, 0)
        selmodel.select(midx, selmodel.ClearAndSelect)

        w.move_up(view)
        d1 = self.get_output(w.Outputs.data, w)
        self.assertEqual(
            d1.domain.attributes,
            data.domain.attributes[:2][::-1] + data.domain.attributes[2:]
        )
        w.move_down(view)
        d1 = self.get_output(w.Outputs.data, w)
        self.assertEqual(
            d1.domain.attributes,
            data.domain.attributes
        )

    def test_drag_drop_move_rows(self):
        data = Table("iris")[:5]
        w = self.widget
        self.send_signal(w.Inputs.data, data)
        used = w.used_attrs_view
        unused = w.available_attrs_view
        model = used.model()
        select_rows(used, [0, 1])

        def drag_exec(self, supported, default):
            mime = self.mimeData()
            drop = QDropEvent(QPointF(20, 20), supported, mime,
                              Qt.NoButton, Qt.NoModifier)
            drop.setDropAction(default)
            drop.setAccepted(False)
            unused.dropEvent(drop)
            assert drop.isAccepted()
            return drop.dropAction()

        with patch.object(QDrag, "exec", drag_exec):
            used.startDrag(Qt.MoveAction)

        self.assertEqual(model.rowCount(), 2)
        self.assertEqual(unused.model().rowCount(), 2)
        out = self.get_output(w.Outputs.data, w)
        self.assertEqual(out.domain.attributes, data.domain.attributes[2:])

    def test_domain_new_feature(self):
        """ Test scenario when new attribute is added at position 0 """
        data = Table("iris")
        self.send_signal(self.widget.Inputs.data, data)

        data1 = Table(
            Domain(
                (ContinuousVariable("a"),) + data.domain.attributes,
                data.domain.class_var),
            np.hstack((np.ones((len(data), 1)), data.X)),
            data.Y
        )
        self.send_signal(self.widget.Inputs.data, data1)

    def test_select_new_features(self):
        """
        When ignore_new_features unchecked new attributes must appear in one of
        selected columns. Test with fist make context remember attributes of
        reduced domain and then testing with full domain. Features in missing
        in reduced domain must appears as seleceted.
        """
        data = Table("iris")
        domain = data.domain

        # data with one feature missing
        new_domain = Domain(
            domain.attributes[:-1], domain.class_var, domain.metas
        )
        new_data = Table.from_table(new_domain, data)

        # make context remember features in reduced domain
        self.send_signal(self.widget.Inputs.data, new_data)
        output = self.get_output(self.widget.Outputs.data)

        self.assertTupleEqual(
            new_data.domain.attributes, output.domain.attributes
        )
        self.assertTupleEqual(new_data.domain.metas, output.domain.metas)
        self.assertEqual(new_data.domain.class_var, output.domain.class_var)

        # send full domain
        self.send_signal(self.widget.Inputs.data, data)
        output = self.get_output(self.widget.Outputs.data)

        # if select_new_features checked all new features goes in the selected
        # features columns - domain equal original
        self.assertFalse(self.widget.ignore_new_features)
        self.assertTupleEqual(data.domain.attributes, output.domain.attributes)
        self.assertTupleEqual(data.domain.metas, output.domain.metas)
        self.assertEqual(data.domain.class_var, output.domain.class_var)

    def test_unselect_new_features(self):
        """
        When ignore_new_features checked new attributes must appear in one
        available attributes column. Test with fist make context remember
        attributes of reduced domain and then testing with full domain.
        Features in missing in reduced domain must appears as not seleceted.
        """
        data = Table("iris")
        domain = data.domain

        # data with one feature missing
        new_domain = Domain(
            domain.attributes[:-1], domain.class_var, domain.metas
        )
        new_data = Table.from_table(new_domain, data)

        # make context remember features in reduced domain
        self.send_signal(self.widget.Inputs.data, new_data)
        # select ignore_new_features
        self.widget.controls.ignore_new_features.click()
        self.assertTrue(self.widget.ignore_new_features)
        output = self.get_output(self.widget.Outputs.data)

        self.assertTupleEqual(
            new_data.domain.attributes, output.domain.attributes
        )
        self.assertTupleEqual(new_data.domain.metas, output.domain.metas)
        self.assertEqual(new_data.domain.class_var, output.domain.class_var)

        # send full domain
        self.send_signal(self.widget.Inputs.data, data)
        output = self.get_output(self.widget.Outputs.data)

        # if ignore_new_features checked all new features goes in the
        # available attributes column
        self.assertTrue(self.widget.ignore_new_features)
        self.assertTupleEqual(new_domain.attributes, output.domain.attributes)
        self.assertTupleEqual(new_domain.metas, output.domain.metas)
        self.assertEqual(new_domain.class_var, output.domain.class_var)
        # test if new attribute was added to unselected attributes
        self.assertEqual(
            domain.attributes[-1], list(self.widget.available_attrs)[0]
        )


if __name__ == "__main__":
    unittest.main()



import ast
from typing import Callable, List, Optional, Union, Dict, Tuple, Any

import numpy as np
from scipy.optimize import curve_fit

from Orange.data import Table, Domain, ContinuousVariable, StringVariable
from Orange.data.filter import HasClass
from Orange.data.util import get_unique_names
from Orange.preprocess import RemoveNaNColumns, Impute
from Orange.regression import Learner, Model

__all__ = ["CurveFitLearner"]


class CurveFitModel(Model):
    def __init__(
            self,
            domain: Domain,
            original_domain: Domain,
            parameters_names: List[str],
            parameters: np.ndarray,
            function: Optional[Callable],
            create_lambda_args: Optional[Tuple]
    ):
        super().__init__(domain, original_domain)
        self.__parameters_names = parameters_names
        self.__parameters = parameters

        if function is None and create_lambda_args is not None:
            function, names, _ = _create_lambda(**create_lambda_args)
            assert parameters_names == names

        assert function

        self.__function = function
        self.__create_lambda_args = create_lambda_args

    @property
    def coefficients(self) -> Table:
        return Table(Domain([ContinuousVariable("coef")],
                            metas=[StringVariable("name")]),
                     self.__parameters[:, None],
                     metas=np.array(self.__parameters_names)[:, None])

    def predict(self, X: np.ndarray) -> np.ndarray:
        predicted = self.__function(X, *self.__parameters)
        if not isinstance(predicted, np.ndarray):
            # handle constant function; i.e. len(self.domain.attributes) == 0
            return np.full(len(X), predicted, dtype=float)
        return predicted.flatten()

    def __getstate__(self) -> Dict:
        if not self.__create_lambda_args:
            raise AttributeError(
                "Can't pickle/copy callable. Use str expression instead."
            )
        return {
            "domain": self.domain,
            "original_domain": self.original_domain,
            "parameters_names": self.__parameters_names,
            "parameters": self.__parameters,
            "function": None,
            "args": self.__create_lambda_args,
        }

    def __setstate__(self, state: Dict):
        self.__init__(*state.values())


class CurveFitLearner(Learner):
    """
    Fit a function to data.
    It uses the scipy.curve_fit to find the optimal values of parameters.

    Parameters
    ----------
    expression : callable or str
        A modeling function.
        If callable, it must take the independent variable as the first
        argument and the parameters to fit as separate remaining arguments.
        If string, a lambda function is created,
        using `expression`, `available_feature_names`, `function` and `env`
        attributes.
        Should be string for pickling the model.
    parameters_names : list of str
        List of parameters names. Only needed when the expression
        is callable.
    features_names : list of str
        List of features names. Only needed when the expression
        is callable.
    available_feature_names : list of str
        List of all available features names. Only needed when the expression
        is string. Needed to distinguish between parameters and features when
        translating the expression into the lambda.
    functions : list of str
        List of all available functions. Only needed when the expression
        is string. Needed to distinguish between parameters and functions when
        translating the expression into the lambda.
    sanitizer : callable
        Function for sanitizing names.
    env : dict
        An environment to capture in the lambda's closure.
    p0 : list of floats, optional
        Initial guess for the parameters.
    bounds : 2-tuple of array_like, optional
        Lower and upper bounds on parameters.
    preprocessors : tuple of Orange preprocessors, optional
        The processors that will be used when data is passed to the learner.

    Examples
    --------
    >>> import numpy as np
    >>> from Orange.data import Table
    >>> from Orange.regression import CurveFitLearner
    >>> data = Table("housing")
    >>> # example with callable expression
    >>> cfun = lambda x, a, b, c: a * np.exp(-b * x[:, 0] * x[:, 1]) + c
    >>> learner = CurveFitLearner(cfun, ["a", "b", "c"], ["CRIM", "LSTAT"])
    >>> model = learner(data)
    >>> pred = model(data)
    >>> coef = model.coefficients
    >>> # example with str expression
    >>> sfun = "a * exp(-b * CRIM * LSTAT) + c"
    >>> names = [a.name for a in data.domain.attributes]
    >>> learner = CurveFitLearner(sfun, available_feature_names=names,
    ...                           functions=["exp"])
    >>> model = learner(data)
    >>> pred = model(data)
    >>> coef = model.coefficients

    """
    preprocessors = [HasClass(), RemoveNaNColumns(), Impute()]
    __returns__ = CurveFitModel
    name = "Curve Fit"

    def __init__(
            self,
            expression: Union[Callable, ast.Expression, str],
            parameters_names: Optional[List[str]] = None,
            features_names: Optional[List[str]] = None,
            available_feature_names: Optional[List[str]] = None,
            functions: Optional[List[str]] = None,
            sanitizer: Optional[Callable] = None,
            env: Optional[Dict[str, Any]] = None,
            p0: Union[List, Dict, None] = None,
            bounds: Union[Tuple, Dict] = (-np.inf, np.inf),
            preprocessors=None
    ):
        super().__init__(preprocessors)

        if callable(expression):
            if parameters_names is None:
                raise TypeError("Provide 'parameters_names' parameter.")
            if features_names is None:
                raise TypeError("Provide 'features_names' parameter.")

            args = None
            function = expression
        else:
            if available_feature_names is None:
                raise TypeError("Provide 'available_feature_names' parameter.")
            if functions is None:
                raise TypeError("Provide 'functions' parameter.")

            args = dict(expression=expression,
                        available_feature_names=available_feature_names,
                        functions=functions, sanitizer=sanitizer, env=env)
            function, parameters_names, features_names = _create_lambda(**args)

        if isinstance(p0, dict):
            p0 = [p0.get(p, 1) for p in parameters_names]
        if isinstance(bounds, dict):
            d = [-np.inf, np.inf]
            lower_bounds = [bounds.get(p, d)[0] for p in parameters_names]
            upper_bounds = [bounds.get(p, d)[1] for p in parameters_names]
            bounds = lower_bounds, upper_bounds

        self.__function = function
        self.__parameters_names = parameters_names
        self.__features_names = features_names
        self.__p0 = p0
        self.__bounds = bounds

        # needed for pickling - if the expression is a lambda function, the
        # learner is not picklable
        self.__create_lambda_args = args

    @property
    def parameters_names(self) -> List[str]:
        return self.__parameters_names

    def fit_storage(self, data: Table) -> CurveFitModel:
        domain: Domain = data.domain
        attributes = []
        for attr in domain.attributes:
            if attr.name in self.__features_names:
                if not attr.is_continuous:
                    raise ValueError("Numeric feature expected.")
                attributes.append(attr)

        new_domain = Domain(attributes, domain.class_vars, domain.metas)
        transformed = data.transform(new_domain)
        params = curve_fit(self.__function, transformed.X, transformed.Y,
                           p0=self.__p0, bounds=self.__bounds)[0]
        return CurveFitModel(new_domain, domain,
                             self.__parameters_names, params, self.__function,
                             self.__create_lambda_args)

    def __getstate__(self) -> Dict:
        if not self.__create_lambda_args:
            raise AttributeError(
                "Can't pickle/copy callable. Use str expression instead."
            )
        state = self.__create_lambda_args.copy()
        state["parameters_names"] = None
        state["features_names"] = None
        state["p0"] = self.__p0
        state["bounds"] = self.__bounds
        state["preprocessors"] = self.preprocessors
        return state

    def __setstate__(self, state: Dict):
        expression = state.pop("expression")
        self.__init__(expression, **state)


def _create_lambda(
        expression: Union[str, ast.Expression] = "",
        available_feature_names: List[str] = None,
        functions: List[str] = None,
        sanitizer: Callable = None,
        env: Optional[Dict[str, Any]] = None
) -> Tuple[Callable, List[str], List[str]]:
    """
    Create a lambda function from a string expression.

    Parameters
    ----------
    expression : str or ast.Expression
        Right side of a modeling function.
    available_feature_names : list of str
        List of all available features names.
        Needed to distinguish between parameters, features and functions.
    functions : list of str
        List of all available functions.
        Needed to distinguish between parameters, features and functions.
    sanitizer : callable, optional
        Function for sanitizing variable names.
    env : dict, optional
        An environment to capture in the lambda's closure.

    Returns
    -------
    func : callable
        The created lambda function.
    params : list of str
        The recognied parameters withint the expression.
    vars_ : list of str
        The recognied variables withint the expression.

    Examples
    --------
    >>> from Orange.data import Table
    >>> data = Table("housing")
    >>> sfun = "a * exp(-b * CRIM * LSTAT) + c"
    >>> names = [a.name for a in data.domain.attributes]
    >>> func, par, var = _create_lambda(sfun, available_feature_names=names,
    ...                                 functions=["exp"], env={"exp": np.exp})
    >>> y = func(data.X, 1, 2, 3)
    >>> par
    ['a', 'b', 'c']
    >>> var
    ['CRIM', 'LSTAT']

    """
    if sanitizer is None:
        sanitizer = lambda n: n
    if env is None:
        env = {name: getattr(np, name) for name in functions}

    exp = ast.parse(expression, mode="eval")
    search = _ParametersSearch(
        [sanitizer(name) for name in available_feature_names],
        functions
    )
    search.visit(exp)
    params = search.parameters
    used_sanitized_feature_names = search.variables

    name = get_unique_names(params, "x")
    feature_mapper = {n: i for i, n in enumerate(used_sanitized_feature_names)}
    exp = _ReplaceVars(name, feature_mapper, functions).visit(exp)

    lambda_ = ast.Lambda(
        args=ast.arguments(
            posonlyargs=[],
            args=[ast.arg(arg=arg) for arg in [name] + params],
            varargs=None,
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[],
        ),
        body=exp.body
    )
    exp = ast.Expression(body=lambda_)
    ast.fix_missing_locations(exp)
    vars_ = [name for name in available_feature_names
             if sanitizer(name) in used_sanitized_feature_names]

    # pylint: disable=eval-used
    return eval(compile(exp, "<lambda>", mode="eval"), env), params, vars_


class _ParametersSearch(ast.NodeVisitor):
    """
    Find features and parameters:
     - feature: if node is instance of ast.Name and is included in vars_names
     - parameters: if node is instance of ast.Name and is not included
     in functions

    Parameters
    ----------
    vars_names : list of str
        List of all available features names.
        Needed to distinguish between parameters, features and functions.
    functions : list of str
        List of all available functions.
        Needed to distinguish between parameters, features and functions.

    Attributes
    ----------
    parameters : list of str
        List of used parameters.
    variables : list of str
        List of used features.

    """

    def __init__(self, vars_names: List[str], functions: List[str]):
        super().__init__()
        self.__vars_names = vars_names
        self.__functions = functions
        self.__parameters: List[str] = []
        self.__variables: List[str] = []

    @property
    def parameters(self) -> List[str]:
        return self.__parameters

    @property
    def variables(self) -> List[str]:
        return self.__variables

    def visit_Name(self, node: ast.Name) -> ast.Name:
        if node.id in self.__vars_names:
            # don't use Set in order to preserve parameters order
            if node.id not in self.__variables:
                self.__variables.append(node.id)
        elif node.id not in self.__functions:
            # don't use Set in order to preserve parameters order
            if node.id not in self.__parameters:
                self.__parameters.append(node.id)
        return node


class _ReplaceVars(ast.NodeTransformer):
    """
    Replace feature names with X[:, i], where i is index of feature.

    Parameters
    ----------
    name : str
        List of all available features names.
        Needed to distinguish between parameters, features and functions.
    vars_mapper : dict
        Dictionary of used features names and the belonging index from domain.
    functions : list of str
        List of all available functions.

    """

    def __init__(self, name: str, vars_mapper: Dict, functions: List):
        super().__init__()
        self.__name = name
        self.__vars_mapper = vars_mapper
        self.__functions = functions

    def visit_Name(self, node: ast.Name) -> Union[ast.Name, ast.Subscript]:
        if node.id not in self.__vars_mapper or node.id in self.__functions:
            return node
        else:
            n = self.__vars_mapper[node.id]
            return ast.Subscript(
                value=ast.Name(id=self.__name, ctx=ast.Load()),
                slice=ast.ExtSlice(
                    dims=[ast.Slice(lower=None, upper=None, step=None),
                          ast.Index(value=ast.Num(n=n))]),
                ctx=node.ctx
            )


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    housing = Table("housing")
    xdata = housing.X
    ydata = housing.Y

    func = lambda x, a, b, c: a * np.exp(-b * x[:, 0]) + c
    pred = CurveFitLearner(func, ["a", "b", "c"], ["LSTAT"])(housing)(housing)

    plt.plot(xdata[:, 12], ydata, "o")
    indices = np.argsort(xdata[:, 12])
    plt.plot(xdata[indices, 12], pred[indices])
    plt.show()



"""Tests for OWPredictions"""
# pylint: disable=protected-access
import os
import unittest
from functools import partial
from tempfile import NamedTemporaryFile
from typing import Optional
from unittest.mock import Mock, patch

import numpy as np

from AnyQt.QtCore import QItemSelectionModel, QItemSelection, Qt, QRect
from AnyQt.QtWidgets import QToolTip

from Orange.base import Model
from Orange.classification import LogisticRegressionLearner, NaiveBayesLearner
from Orange.classification.majority import ConstantModel, MajorityLearner
from Orange.data.io import TabReader
from Orange.evaluation.scoring import TargetScore
from Orange.preprocess import Remove
from Orange.regression import LinearRegressionLearner, MeanLearner
from Orange.widgets.tests.base import WidgetTest, GuiTest
from Orange.widgets.evaluate.owpredictions import (
    OWPredictions, SharedSelectionModel, SharedSelectionStore, DataModel,
    PredictionsModel,
    PredictionsItemDelegate, ClassificationItemDelegate, RegressionItemDelegate,
    NoopItemDelegate, RegressionErrorDelegate, ClassificationErrorDelegate,
    NO_ERR, DIFF_ERROR, ABSDIFF_ERROR, REL_ERROR, ABSREL_ERROR)
from Orange.widgets.evaluate.owcalibrationplot import OWCalibrationPlot
from Orange.widgets.evaluate.owconfusionmatrix import OWConfusionMatrix
from Orange.widgets.evaluate.owliftcurve import OWLiftCurve
from Orange.widgets.evaluate.owrocanalysis import OWROCAnalysis

from Orange.data import Table, Domain, DiscreteVariable, ContinuousVariable
from Orange.modelling import ConstantLearner, TreeLearner
from Orange.evaluation import Results
from Orange.widgets.tests.utils import excepthook_catch, \
    possible_duplicate_table, simulate
from Orange.widgets.utils.colorpalettes import LimitedDiscretePalette


class TestOWPredictions(WidgetTest):

    def setUp(self):
        self.widget = self.create_widget(OWPredictions)  # type: OWPredictions
        self.iris = Table("iris")
        self.iris_classless = self.iris.transform(Domain(self.iris.domain.attributes, []))
        self.housing = Table("housing")

    def test_minimum_size(self):
        pass

    def test_rowCount_from_model(self):
        """Don't crash if the bottom row is visible"""
        self.send_signal(self.widget.Inputs.data, self.iris[:5])
        self.widget.dataview.sizeHintForColumn(0)

    def test_nan_target_input(self):
        data = self.iris[::10].copy()
        with data.unlocked():
            data.Y[1] = np.nan
        yvec = data.get_column(data.domain.class_var)
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.predictors, ConstantLearner()(data), 1)
        pred = self.get_output(self.widget.Outputs.predictions)
        self.assertIsInstance(pred, Table)
        np.testing.assert_array_equal(
            yvec, pred.get_column(data.domain.class_var))

        evres = self.get_output(self.widget.Outputs.evaluation_results)
        self.assertIsInstance(evres, Results)
        self.assertIsInstance(evres.data, Table)
        ev_yvec = evres.data.get_column(data.domain.class_var)

        self.assertTrue(np.all(~np.isnan(ev_yvec)))
        self.assertTrue(np.all(~np.isnan(evres.actual)))

        with data.unlocked():
            data.Y[:] = np.nan
        self.send_signal(self.widget.Inputs.data, data)
        evres = self.get_output(self.widget.Outputs.evaluation_results)
        self.assertEqual(len(evres.data), 0)

    def test_no_values_target(self):
        train = Table("titanic")
        model = ConstantLearner()(train)
        self.send_signal(self.widget.Inputs.predictors, model)
        domain = Domain([DiscreteVariable("status", values=("first", "third")),
                         DiscreteVariable("age", values=("adult", "child")),
                         DiscreteVariable("sex", values=("female", "male"))],
                        [DiscreteVariable("survived", values=())])
        test = Table(domain, np.array([[0, 0, 1], [0, 1, 0], [1, 0, 0]]),
                     np.full((3, 1), np.nan))
        self.send_signal(self.widget.Inputs.data, test)
        pred = self.get_output(self.widget.Outputs.predictions)
        self.assertEqual(len(pred), len(test))

        results = self.get_output(self.widget.Outputs.evaluation_results)

        cm_widget = self.create_widget(OWConfusionMatrix)
        self.send_signal(cm_widget.Inputs.evaluation_results, results,
                         widget=cm_widget)

        ra_widget = self.create_widget(OWROCAnalysis)
        self.send_signal(ra_widget.Inputs.evaluation_results, results,
                         widget=ra_widget)

        lc_widget = self.create_widget(OWLiftCurve)
        self.send_signal(lc_widget.Inputs.evaluation_results, results,
                         widget=lc_widget)

        cp_widget = self.create_widget(OWCalibrationPlot)
        self.send_signal(cp_widget.Inputs.evaluation_results, results,
                         widget=cp_widget)

    def test_mismatching_targets(self):
        warning = self.widget.Warning

        maj_iris = ConstantLearner()(self.iris)
        dom = self.iris.domain
        iris3 = self.iris.transform(Domain(dom[:3], dom[3]))
        maj_iris3 = ConstantLearner()(iris3)

        self.send_signal(self.widget.Inputs.predictors, maj_iris, 1)
        self.send_signal(self.widget.Inputs.predictors, maj_iris3, 2)
        self.assertFalse(warning.wrong_targets.is_shown())

        self.send_signal(self.widget.Inputs.data, self.iris)
        self.assertTrue(warning.wrong_targets.is_shown())

        self.send_signal(self.widget.Inputs.predictors, None, 2)
        self.assertFalse(warning.wrong_targets.is_shown())

        self.send_signal(self.widget.Inputs.predictors, maj_iris3, 2)
        self.assertTrue(warning.wrong_targets.is_shown())

        self.send_signal(self.widget.Inputs.data, None)
        self.assertFalse(warning.wrong_targets.is_shown())

    def test_no_class_on_test(self):
        """Allow test data with no class"""
        titanic = Table("titanic")
        majority_titanic = ConstantLearner()(titanic)

        no_class = titanic.transform(Domain(titanic.domain.attributes, None))
        self.send_signal(self.widget.Inputs.predictors, majority_titanic, 1)
        self.send_signal(self.widget.Inputs.data, no_class)
        out = self.get_output(self.widget.Outputs.predictions)
        np.testing.assert_allclose(out.get_column("constant"), 0)

        predmodel = self.widget.predictionsview.model()
        self.assertTrue(np.isnan(
            predmodel.data(predmodel.index(0, 0), Qt.UserRole)))
        self.assertIn(predmodel.data(predmodel.index(0, 0))[0],
                      titanic.domain.class_var.values)
        self.widget.send_report()

        housing = self.housing[::5]
        mean_housing = ConstantLearner()(housing)
        no_target = housing.transform(Domain(housing.domain.attributes, None))
        self.send_signal(self.widget.Inputs.data, no_target)
        self.send_signal(self.widget.Inputs.predictors, mean_housing, 1)
        self.widget.send_report()

    def test_invalid_regression_target(self):
        widget = self.widget
        self.send_signal(widget.Inputs.predictors,
                         LinearRegressionLearner()(self.housing), 0)

        dom = self.housing.domain
        wrong_class = self.iris.transform(Domain(dom.attributes[:-1],
                                                 dom.attributes[-1]))
        self.send_signal(widget.Inputs.data, wrong_class)

        # can't make a prediction
        predmodel = self.widget.predictionsview.model()
        self.assertTrue(np.isnan(
            predmodel.data(predmodel.index(0, 0), Qt.UserRole)))
        # ... but model reports a value
        self.assertFalse(np.isnan(predmodel.data(predmodel.index(0, 0))[0]))

        no_class = self.iris.transform(Domain(dom.attributes[:-1],
                                              dom.attributes[-1]))
        self.send_signal(widget.Inputs.data, no_class)

        # can't make a prediction
        predmodel = self.widget.predictionsview.model()
        self.assertTrue(np.isnan(
            predmodel.data(predmodel.index(0, 0), Qt.UserRole)))
        # ... but model reports a value
        self.assertFalse(np.isnan(predmodel.data(predmodel.index(0, 0))[0]))

    def test_bad_data(self):
        """
        Firstly it creates predictions with TreeLearner. Then sends predictions and
        different data with different domain to Predictions widget. Those different
        data and domain are similar to original data and domain but they have three
        different target values instead of two.
        GH-2129
        """
        filestr1 = """\
        age\tsex\tsurvived
        d\td\td
        \t\tclass
        adult\tmale\tyes
        adult\tfemale\tno
        child\tmale\tyes
        child\tfemale\tyes
        """
        with NamedTemporaryFile(mode="w", delete=False) as tmp:
            tmp.write(filestr1)
        table = TabReader(tmp.name).read()
        os.unlink(tmp.name)
        learner = TreeLearner()
        tree = learner(table)

        filestr2 = """\
        age\tsex\tsurvived
        d\td\td
        \t\tclass
        adult\tmale\tyes
        adult\tfemale\tno
        child\tmale\tyes
        child\tfemale\tunknown
        """
        with NamedTemporaryFile(mode="w", delete=False) as tmp:
            tmp.write(filestr2)

        bad_table = TabReader(tmp.name).read()
        os.unlink(tmp.name)
        self.send_signal(self.widget.Inputs.predictors, tree, 1)

        with excepthook_catch():
            self.send_signal(self.widget.Inputs.data, bad_table)

    def test_continuous_class(self):
        data = Table("housing")
        cl_data = ConstantLearner()(data)
        self.send_signal(self.widget.Inputs.predictors, cl_data, 1)
        self.send_signal(self.widget.Inputs.data, data)

    def test_changed_class_var(self):
        def set_input(data, model):
            self.send_signals([
                (self.widget.Inputs.data, data),
                (self.widget.Inputs.predictors, model)
            ])

        iris = self.iris
        learner = ConstantLearner()
        heart_disease = Table("heart_disease")
        # catch exceptions in item delegates etc. during switching inputs
        with excepthook_catch():
            set_input(iris[:5], learner(iris))
            set_input(Table("housing"), None)
            set_input(heart_disease[:5], learner(heart_disease))

    def test_predictor_fails(self):
        titanic = Table("titanic")
        failing_model = ConstantLearner()(titanic)
        failing_model.predict = Mock(side_effect=ValueError("foo"))
        self.send_signal(self.widget.Inputs.predictors, failing_model, 1)
        self.send_signal(self.widget.Inputs.data, titanic)

        model2 = ConstantLearner()(titanic)
        self.send_signal(self.widget.Inputs.predictors, model2, 2)

    def test_sort_matching(self):
        def get_items_order(model):
            return model.mapToSourceRows(np.arange(model.rowCount()))

        w = self.widget

        titanic = Table("titanic")
        majority_titanic = LogisticRegressionLearner()(titanic)
        self.send_signal(self.widget.Inputs.predictors, majority_titanic)
        self.send_signal(self.widget.Inputs.data, titanic)

        pred_model = w.predictionsview.model()
        data_model = w.dataview.model()
        n = pred_model.rowCount()

        # no sort
        pred_order = get_items_order(pred_model)
        data_order = get_items_order(data_model)
        np.testing.assert_array_equal(pred_order, np.arange(n))
        np.testing.assert_array_equal(data_order, np.arange(n))

        # sort by first column in prediction table
        pred_model.sort(0)
        w.predictionsview.horizontalHeader().sectionClicked.emit(0)
        pred_order = get_items_order(pred_model)
        data_order = get_items_order(data_model)
        np.testing.assert_array_equal(pred_order, data_order)

        # sort by second column in data table
        data_model.sort(1)
        w.dataview.horizontalHeader().sectionClicked.emit(0)
        pred_order = get_items_order(pred_model)
        data_order = get_items_order(data_model)
        np.testing.assert_array_equal(pred_order, data_order)

        # restore order
        w.reset_button.click()
        pred_order = get_items_order(pred_model)
        data_order = get_items_order(data_model)
        np.testing.assert_array_equal(pred_order, np.arange(n))
        np.testing.assert_array_equal(data_order, np.arange(n))

    def test_sort_predictions(self):
        """
        Test whether sorting of probabilities by FilterSortProxy is correct.
        """

        def get_items_order(model):
            return model.mapToSourceRows(np.arange(model.rowCount()))

        log_reg_iris = LogisticRegressionLearner()(self.iris)
        self.send_signal(self.widget.Inputs.predictors, log_reg_iris)
        self.send_signal(self.widget.Inputs.data, self.iris)
        _, log_reg_probs = log_reg_iris(self.iris, Model.ValueProbs)
        # here I assume that classes are in order 0, 1, 2 in widget

        pred_model = self.widget.predictionsview.model()
        pred_model.sort(0)
        widget_order = get_items_order(pred_model)

        # correct order first sort by probs[:,0] then probs[:,1], ...
        keys = tuple(
            log_reg_probs[:, i] for i in range(
                log_reg_probs.shape[1] - 1, -1, -1))
        sort_order = np.lexsort(keys)
        np.testing.assert_array_equal(widget_order, sort_order)

    def test_reset_no_data(self):
        """
        Check no error when resetting the view without model and data
        """
        self.widget.reset_button.click()

    def test_colors_same_domain(self):
        """
        Test whether the color selection for values is correct.
        """
        # pylint: disable=protected-access
        self.send_signal(self.widget.Inputs.data, self.iris)

        # case 1: one same model
        predictor_iris = ConstantLearner()(self.iris)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris)
        colors = self.widget._get_colors()
        np.testing.assert_array_equal(
            colors, self.iris.domain.class_var.colors)

        # case 2: two same models
        predictor_iris1 = ConstantLearner()(self.iris)
        predictor_iris2 = ConstantLearner()(self.iris)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris1)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris2, 1)
        colors = self.widget._get_colors()
        # assume that colors have same order since values have same order
        np.testing.assert_array_equal(
            colors, self.iris.domain.class_var.colors)

        # case 3: two same models - changed color order
        idom = self.iris.domain
        dom = Domain(
            idom.attributes,
            DiscreteVariable(idom.class_var.name, idom.class_var.values)
        )
        dom.class_var.colors = dom.class_var.colors[::-1]
        iris2 = self.iris.transform(dom)

        predictor_iris1 = ConstantLearner()(iris2)
        predictor_iris2 = ConstantLearner()(iris2)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris1)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris2, 1)
        colors = self.widget._get_colors()
        np.testing.assert_array_equal(colors, iris2.domain.class_var.colors)

    def test_colors_diff_domain(self):
        """
        Test whether the color selection for values is correct.
        """
        # pylint: disable=protected-access
        self.send_signal(self.widget.Inputs.data, self.iris)

        # case 1: two domains one subset other
        idom = self.iris.domain
        dom1 = Domain(
            idom.attributes,
            DiscreteVariable(idom.class_var.name, idom.class_var.values)
        )
        dom2 = Domain(
            idom.attributes,
            DiscreteVariable(idom.class_var.name, idom.class_var.values[:2])
        )
        iris1 = self.iris[:100].transform(dom1)
        iris2 = self.iris[:100].transform(dom2)

        predictor_iris1 = ConstantLearner()(iris1)
        predictor_iris2 = ConstantLearner()(iris2)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris1)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris2, 1)
        colors = self.widget._get_colors()
        np.testing.assert_array_equal(colors, iris1.domain.class_var.colors)

        # case 2: two domains one subset other - different color order
        idom = self.iris.domain
        colors = idom.class_var.colors[::-1]
        dom1 = Domain(
            idom.attributes,
            DiscreteVariable(idom.class_var.name, idom.class_var.values)
        )
        dom2 = Domain(
            idom.attributes,
            DiscreteVariable(idom.class_var.name, idom.class_var.values[:2])
        )
        dom1.class_var.colors = colors
        dom2.class_var.colors = colors[:2]
        iris1 = self.iris[:100].transform(dom1)
        iris2 = self.iris[:100].transform(dom2)

        predictor_iris1 = ConstantLearner()(iris1)
        predictor_iris2 = ConstantLearner()(iris2)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris1)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris2, 1)
        colors = self.widget._get_colors()
        np.testing.assert_array_equal(colors, iris1.domain.class_var.colors)

        # case 3: domain color, values miss-match - use default colors
        idom = self.iris.domain
        dom1 = Domain(
            idom.attributes,
            DiscreteVariable(idom.class_var.name, idom.class_var.values)
        )
        dom2 = Domain(
            idom.attributes,
            DiscreteVariable(idom.class_var.name, idom.class_var.values)
        )
        dom1.class_var.colors = dom1.class_var.colors[::-1]
        iris1 = self.iris.transform(dom1)
        iris2 = self.iris.transform(dom2)

        predictor_iris1 = ConstantLearner()(iris1)
        predictor_iris2 = ConstantLearner()(iris2)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris1)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris2, 1)
        colors = self.widget._get_colors()
        np.testing.assert_array_equal(colors, LimitedDiscretePalette(3).palette)

        # case 4: two domains different values order, matching colors
        idom = self.iris.domain
        # this way we know that default colors are not used
        colors = LimitedDiscretePalette(5).palette[2:]
        dom1 = Domain(
            idom.attributes,
            DiscreteVariable(idom.class_var.name, idom.class_var.values)
        )
        dom2 = Domain(
            idom.attributes,
            DiscreteVariable(idom.class_var.name, idom.class_var.values[::-1])
        )
        dom1.class_var.colors = colors
        dom2.class_var.colors = colors[::-1]  # colors mixed same than values
        iris1 = self.iris[:100].transform(dom1)
        iris2 = self.iris[:100].transform(dom2)

        predictor_iris1 = ConstantLearner()(iris1)
        predictor_iris2 = ConstantLearner()(iris2)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris1)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris2, 1)
        colors = self.widget._get_colors()
        np.testing.assert_array_equal(colors, iris1.domain.class_var.colors)

    def test_colors_continuous(self):
        """
        When only continuous variables in predictor no color should be selected
        we do not work with classes.
        When we add one classifier there should be colors.
        """
        # pylint: disable=protected-access
        data = Table("housing")
        cl_data = ConstantLearner()(data)
        self.send_signal(self.widget.Inputs.predictors, cl_data)
        self.send_signal(self.widget.Inputs.data, data)
        colors = self.widget._get_colors()
        self.assertListEqual([], colors)

        predictor_iris = ConstantLearner()(self.iris)
        self.send_signal(self.widget.Inputs.predictors, predictor_iris, 1)
        colors = self.widget._get_colors()
        self.assertEqual(3, len(colors))

        self.widget.send_report()  # just a quick check that it doesn't crash

    def test_unique_output_domain(self):
        data = possible_duplicate_table('constant')
        predictor = ConstantLearner()(data)
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.predictors, predictor)

        output = self.get_output(self.widget.Outputs.predictions)
        self.assertEqual(output.domain.metas[0].name, 'constant (1)')

    def test_select(self):
        log_reg_iris = LogisticRegressionLearner()(self.iris)
        self.send_signal(self.widget.Inputs.predictors, log_reg_iris)
        self.send_signal(self.widget.Inputs.data, self.iris)

        pred_model = self.widget.predictionsview.model()
        pred_model.sort(0)
        self.widget.predictionsview.selectRow(1)
        sel = {(index.row(), index.column())
               for index in self.widget.dataview.selectionModel().selectedIndexes()}
        self.assertEqual(sel, {(1, col) for col in range(5)})

    def test_select_data_first(self):
        log_reg_iris = LogisticRegressionLearner()(self.iris)
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.send_signal(self.widget.Inputs.predictors, log_reg_iris)

        pred_model = self.widget.predictionsview.model()
        pred_model.sort(0)
        self.widget.predictionsview.selectRow(1)
        sel = {(index.row(), index.column())
               for index in self.widget.dataview.selectionModel().selectedIndexes()}
        self.assertEqual(sel, {(1, col) for col in range(5)})

    def test_selection_in_setting(self):
        widget = self.create_widget(OWPredictions,
                                    stored_settings={"selection": [1, 3, 4]})
        self.send_signal(widget.Inputs.data, self.iris)
        log_reg_iris = LogisticRegressionLearner()(self.iris)
        self.send_signal(widget.Inputs.predictors, log_reg_iris)
        sel = {(index.row(), index.column())
               for index in widget.dataview.selectionModel().selectedIndexes()}
        self.assertEqual(sel, {(row, col)
                               for row in [1, 3, 4] for col in range(5)})
        out = self.get_output(widget.Outputs.predictions)
        exp = self.iris[np.array([1, 3, 4])]
        np.testing.assert_equal(out.X, exp.X)

    def test_unregister_prediction_model(self):
        log_reg_iris = LogisticRegressionLearner()(self.iris)
        self.send_signal(self.widget.Inputs.predictors, log_reg_iris)
        self.send_signal(self.widget.Inputs.data, self.iris)
        self.widget.selection_store.unregister = Mock()
        prev_model = self.widget.predictionsview.model()
        self.send_signal(self.widget.Inputs.predictors, log_reg_iris)
        self.widget.selection_store.unregister.called_with(prev_model)

    def test_multi_inputs(self):
        w = self.widget
        data = self.iris[::5].copy()

        p1 = ConstantLearner()(data)
        p1.name = "P1"
        p2 = ConstantLearner()(data)
        p2.name = "P2"
        p3 = ConstantLearner()(data)
        p3.name = "P3"
        for i, p in enumerate([p1, p2, p3], 1):
            self.send_signal(w.Inputs.predictors, p, i)
        self.send_signal(w.Inputs.data, data)

        def check_evres(expected):
            out = self.get_output(w.Outputs.evaluation_results)
            self.assertSequenceEqual(out.learner_names, expected)
            self.assertEqual(out.folds, [...])
            self.assertEqual(out.models.shape, (1, len(out.learner_names)))
            self.assertIsInstance(out.models[0, 0], ConstantModel)

        check_evres(["P1", "P2", "P3"])

        self.send_signal(w.Inputs.predictors, None, 2)
        check_evres(["P1", "P3"])

        self.send_signal(w.Inputs.predictors, p2, 2)
        check_evres(["P1", "P2", "P3"])

        self.send_signal(w.Inputs.predictors,
                         w.Inputs.predictors.closing_sentinel, 2)
        check_evres(["P1", "P3"])

        self.send_signal(w.Inputs.predictors, p2, 2)
        check_evres(["P1", "P3", "P2"])

    def test_missing_target_cls(self):
        mask = np.zeros(len(self.iris), dtype=bool)
        mask[::2] = True
        train_data = self.iris[~mask]
        predict_data = self.iris[mask]
        model = LogisticRegressionLearner()(train_data)

        self.send_signal(self.widget.Inputs.predictors, model)
        self.send_signal(self.widget.Inputs.data, predict_data)
        self.assertFalse(self.widget.Warning.missing_targets.is_shown())
        self.assertFalse(self.widget.Error.scorer_failed.is_shown())

        with predict_data.unlocked():
            predict_data.Y[0] = np.nan
        self.send_signal(self.widget.Inputs.data, predict_data)
        self.assertTrue(self.widget.Warning.missing_targets.is_shown())
        self.assertFalse(self.widget.Error.scorer_failed.is_shown())

        with predict_data.unlocked():
            predict_data.Y[:] = np.nan
        self.send_signal(self.widget.Inputs.data, predict_data)
        self.assertTrue(self.widget.Warning.missing_targets.is_shown())
        self.assertFalse(self.widget.Error.scorer_failed.is_shown())

        self.send_signal(self.widget.Inputs.predictors, None)
        self.assertFalse(self.widget.Warning.missing_targets.is_shown())
        self.assertFalse(self.widget.Error.scorer_failed.is_shown())

    def test_missing_target_reg(self):
        mask = np.zeros(len(self.housing), dtype=bool)
        mask[::2] = True
        train_data = self.housing[~mask]
        predict_data = self.housing[mask]
        model = LinearRegressionLearner()(train_data)

        self.send_signal(self.widget.Inputs.predictors, model)
        self.send_signal(self.widget.Inputs.data, predict_data)
        self.assertFalse(self.widget.Warning.missing_targets.is_shown())
        self.assertFalse(self.widget.Error.scorer_failed.is_shown())

        with predict_data.unlocked():
            predict_data.Y[0] = np.nan
        self.send_signal(self.widget.Inputs.data, predict_data)
        self.assertTrue(self.widget.Warning.missing_targets.is_shown())
        self.assertFalse(self.widget.Error.scorer_failed.is_shown())

        with predict_data.unlocked():
            predict_data.Y[:] = np.nan
        self.send_signal(self.widget.Inputs.data, predict_data)
        self.assertTrue(self.widget.Warning.missing_targets.is_shown())
        self.assertFalse(self.widget.Error.scorer_failed.is_shown())

        self.send_signal(self.widget.Inputs.predictors, None)
        self.assertFalse(self.widget.Warning.missing_targets.is_shown())
        self.assertFalse(self.widget.Error.scorer_failed.is_shown())

    def _mock_predictors(self):
        def pred(values):
            slot = Mock()
            slot.predictor.domain = Domain([], DiscreteVariable("c", tuple(values)))
            return slot

        def predc():
            slot = Mock()
            slot.predictor.domain = Domain([], ContinuousVariable("c"))
            return slot

        widget = self.widget
        model = Mock()
        model.setProbInd = Mock()
        widget.predictionsview.model = Mock(return_value=model)

        widget.predictors = \
            [pred(values) for values in ("abc", "ab", "cbd", "e")] + [predc()]

    def test_update_prediction_delegate_discrete(self):
        self._mock_predictors()

        widget = self.widget
        prob_combo = widget.controls.shown_probs
        set_prob_ind = widget.predictionsview.model().setProbInd
        widget._non_errored_predictors = lambda: widget.predictors[:4]

        widget.data = Table.from_list(
            Domain([], DiscreteVariable("c", values=tuple("abc"))), [])

        widget._update_control_visibility()
        self.assertFalse(prob_combo.isHidden())

        widget._set_class_values()
        self.assertEqual(widget.class_values, list("abcde"))

        widget._set_target_combos()
        self.assertEqual(
            [prob_combo.itemText(i) for i in range(prob_combo.count())],
            widget.PROB_OPTS + list("abc"))

        widget.shown_probs = widget.NO_PROBS
        widget._update_prediction_delegate()
        for delegate in widget._delegates:
            if isinstance(delegate, ClassificationItemDelegate):
                self.assertEqual(list(delegate.shown_probabilities), [])
                self.assertEqual(delegate.tooltip, "")
        set_prob_ind.assert_called_with([[], [], [], []])

        widget.shown_probs = widget.DATA_PROBS
        widget._update_prediction_delegate()
        self.assertEqual(widget._delegates[0].shown_probabilities, [0, 1, 2])
        self.assertEqual(widget._delegates[2].shown_probabilities, [0, 1, None])
        self.assertEqual(widget._delegates[2].shown_probabilities, [0, 1, None])
        self.assertEqual(widget._delegates[4].shown_probabilities, [None, 1, 2])
        self.assertEqual(widget._delegates[6].shown_probabilities, [None, None, None])
        for delegate in widget._delegates[:-1:2]:
            self.assertEqual(delegate.tooltip, "p(a, b, c)")
        set_prob_ind.assert_called_with([[0, 1, 2], [0, 1], [1, 2], []])

        widget.shown_probs = widget.MODEL_PROBS
        widget._update_prediction_delegate()
        self.assertEqual(widget._delegates[0].shown_probabilities, [0, 1, 2])
        self.assertEqual(widget._delegates[0].tooltip, "p(a, b, c)")
        self.assertEqual(widget._delegates[2].shown_probabilities, [0, 1])
        self.assertEqual(widget._delegates[2].tooltip, "p(a, b)")
        self.assertEqual(widget._delegates[4].shown_probabilities, [2, 1, 3])
        self.assertEqual(widget._delegates[4].tooltip, "p(c, b, d)")
        self.assertEqual(widget._delegates[6].shown_probabilities, [4])
        self.assertEqual(widget._delegates[6].tooltip, "p(e)")
        set_prob_ind.assert_called_with([[0, 1, 2], [0, 1], [2, 1, 3], [4]])

        widget.shown_probs = widget.BOTH_PROBS
        widget._update_prediction_delegate()
        self.assertEqual(widget._delegates[0].shown_probabilities, [0, 1, 2])
        self.assertEqual(widget._delegates[0].tooltip, "p(a, b, c)")
        self.assertEqual(widget._delegates[2].shown_probabilities, [0, 1])
        self.assertEqual(widget._delegates[2].tooltip, "p(a, b)")
        self.assertEqual(widget._delegates[4].shown_probabilities, [1, 2])
        self.assertEqual(widget._delegates[4].tooltip, "p(b, c)")
        self.assertEqual(widget._delegates[6].shown_probabilities, [])
        self.assertEqual(widget._delegates[6].tooltip, "")
        set_prob_ind.assert_called_with([[0, 1, 2], [0, 1], [1, 2], []])

        n_fixed = len(widget.PROB_OPTS)
        widget.shown_probs = n_fixed  # a
        widget._update_prediction_delegate()
        self.assertEqual(widget._delegates[0].shown_probabilities, [0])
        self.assertEqual(widget._delegates[2].shown_probabilities, [0])
        self.assertEqual(widget._delegates[4].shown_probabilities, [None])
        self.assertEqual(widget._delegates[6].shown_probabilities, [None])
        for delegate in widget._delegates[:-1:2]:
            self.assertEqual(delegate.tooltip, "p(a)")
        set_prob_ind.assert_called_with([[0], [0], [], []])

        n_fixed = len(widget.PROB_OPTS)
        widget.shown_probs = n_fixed + 1  # b
        widget._update_prediction_delegate()
        self.assertEqual(widget._delegates[0].shown_probabilities, [1])
        self.assertEqual(widget._delegates[2].shown_probabilities, [1])
        self.assertEqual(widget._delegates[4].shown_probabilities, [1])
        self.assertEqual(widget._delegates[6].shown_probabilities, [None])
        for delegate in widget._delegates[:-1:2]:
            self.assertEqual(delegate.tooltip, "p(b)")
        set_prob_ind.assert_called_with([[1], [1], [1], []])

        n_fixed = len(widget.PROB_OPTS)
        widget.shown_probs = n_fixed + 2  # c
        widget._update_prediction_delegate()
        self.assertEqual(widget._delegates[0].shown_probabilities, [2])
        self.assertEqual(widget._delegates[2].shown_probabilities, [None])
        self.assertEqual(widget._delegates[4].shown_probabilities, [2])
        self.assertEqual(widget._delegates[6].shown_probabilities, [None])
        for delegate in widget._delegates[:-1:2]:
            self.assertEqual(delegate.tooltip, "p(c)")
        set_prob_ind.assert_called_with([[2], [], [2], []])

    def test_update_delegates_continuous(self):
        self._mock_predictors()

        widget = self.widget
        widget.shown_probs = widget.DATA_PROBS

        widget.data = Table.from_list(Domain([], ContinuousVariable("c")), [])

        # only regression
        all_predictors = widget.predictors
        widget.predictors = [widget.predictors[-1]]
        widget._update_control_visibility()
        self.assertTrue(widget.controls.shown_probs.isHidden())
        self.assertTrue(widget.controls.target_class.isHidden())

        # regression and classification
        widget.predictors = all_predictors
        widget._update_control_visibility()
        self.assertFalse(widget.controls.shown_probs.isHidden())
        self.assertTrue(widget.controls.target_class.isHidden())

        widget._set_class_values()
        self.assertEqual(widget.class_values, list("abcde"))

        widget._set_target_combos()
        self.assertEqual(widget.shown_probs, widget.NO_PROBS)

        def is_enabled(prob_item):
            return widget.controls.shown_probs.model().item(prob_item).flags() & Qt.ItemIsEnabled
        self.assertTrue(is_enabled(widget.NO_PROBS))
        self.assertTrue(is_enabled(widget.MODEL_PROBS))
        self.assertFalse(is_enabled(widget.DATA_PROBS))
        self.assertFalse(is_enabled(widget.BOTH_PROBS))

    def test_delegate_ranges(self):
        widget = self.widget

        class Model1(Model):
            name = "foo"

            def predict(self, X):
                return X[:, 0] - 2

        class Model2(Model):
            name = "bar"

            def predict(self, X):
                return np.full(len(X), np.nan)

        domain = Domain([ContinuousVariable("x")], ContinuousVariable("y"))
        x = np.arange(12, 17, dtype=float)[:, None]
        y = np.array([12, 13, 14, 15, np.nan])
        data = Table(domain, x, y)

        ddomain = Domain(
            [ContinuousVariable("x")],
            DiscreteVariable("y", values=tuple("abcdefghijklmnopq")))
        self.send_signal(widget.Inputs.data, data)
        self.send_signal(widget.Inputs.predictors, Model1(domain), 1)
        self.send_signal(widget.Inputs.predictors, Model2(domain), 2)
        self.send_signal(widget.Inputs.predictors, Model1(ddomain), 3)

        delegate = widget.predictionsview.itemDelegateForColumn(0)
        # values for model are 10 to 14 (incl), Y goes from 12 to 15 (incl)
        self.assertIsInstance(delegate, RegressionItemDelegate)
        self.assertEqual(delegate.offset, 10)
        self.assertEqual(delegate.span, 5)

        delegate = widget.predictionsview.itemDelegateForColumn(2)
        # values for model are all-nan, Y goes from 12 to 15 (incl)
        self.assertIsInstance(delegate, RegressionItemDelegate)
        self.assertEqual(delegate.offset, 12)
        self.assertEqual(delegate.span, 3)

        delegate = widget.predictionsview.itemDelegateForColumn(4)
        self.assertIsInstance(delegate, ClassificationItemDelegate)

        data = Table(domain, x, np.full(5, np.nan))
        self.send_signal(widget.Inputs.data, data)
        delegate = widget.predictionsview.itemDelegateForColumn(0)
        # values for model are 10 to 14 (incl), Y is nan
        self.assertIsInstance(delegate, RegressionItemDelegate)
        self.assertEqual(delegate.offset, 10)
        self.assertEqual(delegate.span, 4)

        delegate = widget.predictionsview.itemDelegateForColumn(2)
        # values for model and y are nan
        self.assertIsInstance(delegate, RegressionItemDelegate)
        self.assertEqual(delegate.offset, 0)
        self.assertEqual(delegate.span, 1)

        delegate = widget.predictionsview.itemDelegateForColumn(4)
        self.assertIsInstance(delegate, ClassificationItemDelegate)

    class _Scorer(TargetScore):
        # pylint: disable=arguments-differ
        def compute_score(self, _, target, **__):
            return [42 if target is None else target]

    def test_output_wrt_shown_probs_1(self):
        """Data has one class less, models have same, different or one more"""
        widget = self.widget
        iris012 = self.iris
        purge = Remove(class_flags=Remove.RemoveUnusedValues)
        iris01 = purge(iris012[:100])
        iris12 = purge(iris012[50:])

        bayes01 = NaiveBayesLearner()(iris01)
        bayes12 = NaiveBayesLearner()(iris12)
        bayes012 = NaiveBayesLearner()(iris012)

        self.send_signal(widget.Inputs.data, iris01)
        self.send_signal(widget.Inputs.predictors, bayes01, 0)
        self.send_signal(widget.Inputs.predictors, bayes12, 1)
        self.send_signal(widget.Inputs.predictors, bayes012, 2)
        widget.controls.show_probability_errors.setChecked(False)

        for i, pred in enumerate(widget.predictors):
            p = pred.results.unmapped_probabilities
            p[0] = 10 + 100 * i + np.arange(p.shape[1])
            pred.results.unmapped_predicted[:] = i

        widget.shown_probs = widget.NO_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 1, 2])

        widget.shown_probs = widget.DATA_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 10, 11, 1, 0, 110, 2, 210, 211])

        widget.shown_probs = widget.MODEL_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 10, 11, 1, 110, 111, 2, 210, 211, 212])

        widget.shown_probs = widget.BOTH_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 10, 11, 1, 110, 2, 210, 211])

        widget.shown_probs = widget.BOTH_PROBS + 1
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 10, 1, 0, 2, 210])

        widget.shown_probs = widget.BOTH_PROBS + 2
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 11, 1, 110, 2, 211])

    def test_output_wrt_shown_probs_2(self):
        """One model misses one class"""
        widget = self.widget
        iris012 = self.iris
        purge = Remove(class_flags=Remove.RemoveUnusedValues)
        iris01 = purge(iris012[:100])

        bayes01 = NaiveBayesLearner()(iris01)
        bayes012 = NaiveBayesLearner()(iris012)

        self.send_signal(widget.Inputs.data, iris012)
        self.send_signal(widget.Inputs.predictors, bayes01, 0)
        self.send_signal(widget.Inputs.predictors, bayes012, 1)
        widget.controls.show_probability_errors.setChecked(False)

        for i, pred in enumerate(widget.predictors):
            p = pred.results.unmapped_probabilities
            p[0] = 10 + 100 * i + np.arange(p.shape[1])
            pred.results.unmapped_predicted[:] = i

        widget.shown_probs = widget.NO_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 1])

        widget.shown_probs = widget.DATA_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 10, 11, 0, 1, 110, 111, 112])

        widget.shown_probs = widget.MODEL_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 10, 11, 1, 110, 111, 112])

        widget.shown_probs = widget.BOTH_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 10, 11, 1, 110, 111, 112])

        widget.shown_probs = widget.BOTH_PROBS + 1
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 10, 1, 110])

        widget.shown_probs = widget.BOTH_PROBS + 2
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 11, 1, 111])

        widget.shown_probs = widget.BOTH_PROBS + 3
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 0, 1, 112])

    def test_output_regression(self):
        widget = self.widget
        self.send_signal(widget.Inputs.data, self.housing)
        self.send_signal(widget.Inputs.predictors,
                         LinearRegressionLearner()(self.housing), 0)
        self.send_signal(widget.Inputs.predictors,
                         MeanLearner()(self.housing), 1)
        out = self.get_output(widget.Outputs.predictions)
        np.testing.assert_equal(
            out.metas[:, [0, 2]],
            np.hstack([pred.results.predicted.T for pred in widget.predictors]))

    def test_classless(self):
        widget = self.widget
        iris012 = self.iris
        purge = Remove(class_flags=Remove.RemoveUnusedValues)
        iris01 = purge(iris012[:100])
        iris12 = purge(iris012[50:])

        bayes01 = NaiveBayesLearner()(iris01)
        bayes12 = NaiveBayesLearner()(iris12)
        bayes012 = NaiveBayesLearner()(iris012)

        self.send_signal(widget.Inputs.data, self.iris_classless)
        self.send_signal(widget.Inputs.predictors, bayes01, 0)
        self.send_signal(widget.Inputs.predictors, bayes12, 1)
        self.send_signal(widget.Inputs.predictors, bayes012, 2)

        for i, pred in enumerate(widget.predictors):
            p = pred.results.unmapped_probabilities
            p[0] = 10 + 100 * i + np.arange(p.shape[1])
            pred.results.unmapped_predicted[:] = i

        widget.shown_probs = widget.NO_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 1, 2])

        widget.shown_probs = widget.MODEL_PROBS
        widget._commit_predictions()
        out = self.get_output(widget.Outputs.predictions)
        self.assertEqual(list(out.metas[0]), [0, 10, 11, 1, 110, 111, 2, 210, 211, 212])

    @patch("Orange.widgets.evaluate.owpredictions.usable_scorers",
           Mock(return_value=[_Scorer]))
    def test_change_target(self):
        widget = self.widget
        table = widget.score_table
        combo = widget.controls.target_class

        log_reg_iris = LogisticRegressionLearner()(self.iris)
        self.send_signal(widget.Inputs.predictors, log_reg_iris)
        self.send_signal(widget.Inputs.data, self.iris)

        self.assertEqual(table.model.rowCount(), 1)
        self.assertEqual(table.model.columnCount(), 4)
        self.assertEqual(float(table.model.data(table.model.index(0, 3))), 42)

        for idx, value in enumerate(widget.class_var.values):
            simulate.combobox_activate_item(combo, value, Qt.DisplayRole)
            self.assertEqual(table.model.rowCount(), 1)
            self.assertEqual(table.model.columnCount(), 4)
            self.assertEqual(float(table.model.data(table.model.index(0, 3))),
                             idx)

    def test_multi_target_input(self):
        widget = self.widget

        domain = Domain([ContinuousVariable('var1')],
                        class_vars=[
                            ContinuousVariable('c1'),
                            DiscreteVariable('c2', values=('no', 'yes'))
                        ])
        data = Table.from_list(domain, [[1, 5, 0], [2, 10, 1]])

        mock_model = Mock(spec=Model, return_value=np.asarray([0.2, 0.1]))
        mock_model.name = 'Mockery'
        mock_model.domain = domain

        self.send_signal(widget.Inputs.data, data)
        self.send_signal(widget.Inputs.predictors, mock_model, 1)
        pred = self.get_output(widget.Outputs.predictions)
        self.assertIsInstance(pred, Table)

    def test_error_controls_visibility(self):
        widget = self.widget
        senddata = partial(self.send_signal, widget.Inputs.data)
        sendpredictor = partial(self.send_signal, widget.Inputs.predictors)
        clshidden = widget._cls_error_controls[0].isHidden
        reghidden = widget._reg_error_controls[0].isHidden
        colhidden = widget.predictionsview.isColumnHidden
        delegate = widget.predictionsview.itemDelegateForColumn

        iris = self.iris
        regiris = iris.transform(Domain(iris.domain.attributes[:3],
                                        iris.domain.attributes[3]))
        riris = MeanLearner()(regiris)
        ciris = MajorityLearner()(iris)

        self.assertFalse(clshidden())
        self.assertFalse(reghidden())

        senddata(self.housing)
        self.assertTrue(clshidden())
        self.assertFalse(reghidden())

        senddata(self.iris)
        self.assertFalse(clshidden())
        self.assertTrue(reghidden())

        senddata(None)
        self.assertTrue(clshidden())
        self.assertTrue(reghidden())

        senddata(self.iris_classless)
        self.assertTrue(clshidden())
        self.assertTrue(reghidden())

        sendpredictor(ciris, 0)
        sendpredictor(riris, 1)
        self.assertFalse(colhidden(0))
        self.assertTrue(colhidden(1))
        self.assertFalse(colhidden(2))
        self.assertTrue(colhidden(3))
        self.assertIsInstance(delegate(1), NoopItemDelegate)
        self.assertIsInstance(delegate(3), NoopItemDelegate)

        senddata(regiris)
        self.assertFalse(colhidden(0))
        self.assertTrue(colhidden(1))
        self.assertFalse(colhidden(2))
        self.assertFalse(colhidden(3))
        self.assertIsInstance(delegate(1), NoopItemDelegate)
        self.assertIsInstance(delegate(3), RegressionErrorDelegate)

        err_combo = self.widget.controls.show_reg_errors
        err_combo.setCurrentIndex(0)
        err_combo.activated.emit(0)
        self.assertTrue(colhidden(1))
        self.assertTrue(colhidden(3))
        self.assertIsInstance(delegate(1), NoopItemDelegate)
        self.assertIsInstance(delegate(3), (RegressionErrorDelegate,
                                            NoopItemDelegate))

        senddata(iris)
        self.assertFalse(colhidden(1))
        self.assertTrue(colhidden(3))
        self.assertIsInstance(delegate(1), ClassificationErrorDelegate)
        self.assertIsInstance(delegate(3), NoopItemDelegate)

        err_box = self.widget.controls.show_probability_errors
        err_box.click()
        self.assertTrue(colhidden(1))
        self.assertIsInstance(delegate(1), (ClassificationErrorDelegate,
                                            NoopItemDelegate))
        self.assertIsInstance(delegate(3), NoopItemDelegate)

    def test_regression_error_delegate_ranges(self):
        def set_type(tpe):
            combo = widget.controls.show_reg_errors
            combo.setCurrentIndex(tpe)
            combo.activated.emit(tpe)

        def get_delegate() -> Optional[RegressionErrorDelegate]:
            return widget.predictionsview.itemDelegateForColumn(1)

        widget = self.widget
        domain = Domain([ContinuousVariable("x")],
                        ContinuousVariable("y"))
        data = Table.from_numpy(domain, np.arange(2, 12)[:, None], np.arange(2, 12))
        model = MeanLearner()(data)
        model.mean = 5
        self.send_signal(widget.Inputs.data, data)
        self.send_signal(widget.Inputs.predictors, model, 0)

        set_type(NO_ERR)
        self.assertIsInstance(get_delegate(), NoopItemDelegate)

        set_type(DIFF_ERROR)
        delegate = get_delegate()
        self.assertEqual(delegate.span, 6)
        self.assertTrue(delegate.centered)

        set_type(ABSDIFF_ERROR)
        delegate = get_delegate()
        self.assertEqual(delegate.span, 6)
        self.assertFalse(delegate.centered)

        set_type(REL_ERROR)
        delegate = get_delegate()
        self.assertEqual(delegate.span, max(3 / 2, 6 / 11))
        self.assertTrue(delegate.centered)

        set_type(ABSREL_ERROR)
        delegate = get_delegate()
        self.assertEqual(delegate.span, max(3 / 2, 6 / 11))
        self.assertFalse(delegate.centered)

    def test_regression_error_no_model(self):
        data = self.housing[:5]
        self.send_signal(self.widget.Inputs.data, data)
        combo = self.widget.controls.show_reg_errors
        with excepthook_catch(raise_on_exit=True):
            simulate.combobox_activate_index(combo, 1)

    def test_report(self):
        widget = self.widget

        log_reg_iris = LogisticRegressionLearner()(self.iris)
        self.send_signal(widget.Inputs.predictors, log_reg_iris)
        self.send_signal(widget.Inputs.data, self.iris)

        widget.report_paragraph = Mock()
        reports = set()
        for widget.shown_probs in range(len(widget.PROB_OPTS)):
            widget.send_report()
            reports.add(widget.report_paragraph.call_args[0][1])
        self.assertEqual(len(reports), len(widget.PROB_OPTS))

        for widget.shown_probs, value in enumerate(
                widget.class_var.values, start=widget.shown_probs + 1):
            widget.send_report()
            self.assertIn(value, widget.report_paragraph.call_args[0][1])

    def test_migrate_shown_scores(self):
        settings = {"score_table": {"shown_scores": {"Sensitivity"}}}
        self.widget.migrate_settings(settings, 1)
        self.assertTrue(settings["score_table"]["show_score_hints"]["Sensitivity"])

    def test_output_error_reg(self):
        data = self.housing
        lin_reg = LinearRegressionLearner()
        self.send_signal(self.widget.Inputs.data, data)
        self.send_signal(self.widget.Inputs.predictors, lin_reg(data), 0)
        self.send_signal(self.widget.Inputs.predictors,
                         LinearRegressionLearner(fit_intercept=False)(data), 1)
        pred = self.get_output(self.widget.Outputs.predictions)

        names = ["", " (error)"]
        names = [f"{n}{i}" for i in ("", " (1)") for n in names]
        names = [f"{lin_reg.name}{x}" for x in names]
        self.assertEqual(names, [m.name for m in pred.domain.metas])
        self.assertAlmostEqual(pred.metas[0, 1], 6.0, 1)
        self.assertAlmostEqual(pred.metas[0, 3], 5.1, 1)

    def test_output_error_cls(self):
        data = self.iris
        log_reg = LogisticRegressionLearner()
        self.send_signal(self.widget.Inputs.predictors, log_reg(data), 0)
        self.send_signal(self.widget.Inputs.predictors,
                         LogisticRegressionLearner(penalty="l1")(data), 1)
        with data.unlocked(data.Y):
            data.Y[1] = np.nan
        self.send_signal(self.widget.Inputs.data, data)
        pred = self.get_output(self.widget.Outputs.predictions)

        names = [""] + [f" ({v})" for v in
                        list(data.domain.class_var.values) + ["error"]]
        names = [f"{n}{i}" for i in ("", " (1)") for n in names]
        names = [f"{log_reg.name}{x}" for x in names]
        self.assertEqual(names, [m.name for m in pred.domain.metas])
        self.assertAlmostEqual(pred.metas[0, 4], 0.018, 3)
        self.assertAlmostEqual(pred.metas[0, 9], 0.113, 3)
        self.assertTrue(np.isnan(pred.metas[1, 4]))
        self.assertTrue(np.isnan(pred.metas[1, 9]))


class SelectionModelTest(unittest.TestCase):
    def setUp(self):
        iris = Table("iris")

        self.datamodel1 = DataModel(iris[:5, :2])
        self.store = SharedSelectionStore(self.datamodel1)
        self.model1 = SharedSelectionModel(self.store, self.datamodel1, None)

        self.datamodel2 = DataModel(iris[-5:, :3])
        self.model2 = SharedSelectionModel(self.store, self.datamodel2, None)

    def itsel(self, rows):
        sel = QItemSelection()
        for row in rows:
            index = self.store.model.index(row, 0)
            sel.select(index, index)
        return sel


class SharedSelectionStoreTest(SelectionModelTest):
    def test_registration(self):
        self.store.unregister(self.model1)
        self.store.unregister(self.model1)  # should have no effect and no error
        self.store.unregister(self.model2)

    def test_select_rows(self):
        emit1 = self.model1.emit_selection_rows_changed = Mock()
        emit2 = self.model2.emit_selection_rows_changed = Mock()

        store = self.store
        store.select_rows({1, 2}, QItemSelectionModel.Select)
        self.assertEqual(store.rows, {1, 2})
        emit1.assert_called_with([1, 2], [])
        emit2.assert_called_with([1, 2], [])

        store.select_rows({1, 2, 4}, QItemSelectionModel.Select)
        self.assertEqual(store.rows, {1, 2, 4})
        emit1.assert_called_with([4], [])
        emit2.assert_called_with([4], [])

        store.select_rows({3, 4}, QItemSelectionModel.Toggle)
        self.assertEqual(store.rows, {1, 2, 3})
        emit1.assert_called_with([3], [4])
        emit2.assert_called_with([3], [4])

        store.select_rows({0, 2}, QItemSelectionModel.Deselect)
        self.assertEqual(store.rows, {1, 3})
        emit1.assert_called_with([], [2])
        emit2.assert_called_with([], [2])

        store.select_rows({2, 3, 4}, QItemSelectionModel.ClearAndSelect)
        self.assertEqual(store.rows, {2, 3, 4})
        emit1.assert_called_with([2, 4], [1])
        emit2.assert_called_with([2, 4], [1])

        store.select_rows({2, 3, 4}, QItemSelectionModel.Clear)
        self.assertEqual(store.rows, set())
        emit1.assert_called_with([], [2, 3, 4])
        emit2.assert_called_with([], [2, 3, 4])

        store.select_rows({2, 3, 4}, QItemSelectionModel.ClearAndSelect)
        emit1.reset_mock()
        emit2.reset_mock()
        store.select_rows({2, 4}, QItemSelectionModel.Select)
        emit1.assert_not_called()
        emit2.assert_not_called()

    def test_select_maps_from_proxy(self):
        store = self.store
        store.select_rows = Mock()

        # Map QItemSelection
        store.model.setSortIndices(np.array([4, 0, 1, 2, 3]))
        store.select(self.itsel([1, 2]), QItemSelectionModel.Select)
        store.select_rows.assert_called_with({0, 1}, QItemSelectionModel.Select)

        # Map QModelIndex
        store.model.setSortIndices(np.array([4, 0, 1, 2, 3]))
        store.select(store.model.index(0, 0), QItemSelectionModel.Select)
        store.select_rows.assert_called_with({4}, QItemSelectionModel.Select)

        # Map empty selection
        store.select(self.itsel([]), QItemSelectionModel.Select)
        store.select_rows.assert_called_with(set(), QItemSelectionModel.Select)

    def test_clear(self):
        store = self.store

        store.select_rows({1, 2, 4}, QItemSelectionModel.Select)
        self.assertEqual(store.rows, {1, 2, 4})

        emit1 = self.model1.emit_selection_rows_changed = Mock()
        emit2 = self.model2.emit_selection_rows_changed = Mock()

        store.clear_selection()
        self.assertEqual(store.rows, set())
        emit1.assert_called_with([], [1, 2, 4])
        emit2.assert_called_with([], [1, 2, 4])

    def test_reset(self):
        store = self.store

        store.select_rows({1, 2, 4}, QItemSelectionModel.Select)
        self.assertEqual(store.rows, {1, 2, 4})

        emit1 = self.model1.emit_selection_rows_changed = Mock()
        emit2 = self.model2.emit_selection_rows_changed = Mock()

        store.reset()
        self.assertEqual(store.rows, set())
        emit1.assert_not_called()
        emit2.assert_not_called()

    def test_emit_changed_maps_to_proxy(self):
        store = self.store
        emit1 = self.model1.emit_selection_rows_changed = Mock()
        self.model2.emit_selection_rows_changed = Mock()

        def assert_called(exp_selected, exp_deselected):
            # pylint: disable=unsubscriptable-object
            selected, deselected = emit1.call_args[0]
            self.assertEqual(list(selected), exp_selected)
            self.assertEqual(list(deselected), exp_deselected)

        store.model.setSortIndices([4, 0, 1, 2, 3])
        store.select_rows({3, 4}, QItemSelectionModel.Select)
        assert_called([4, 0], [])

        store.model.setSortIndices(None)
        store.select_rows({4}, QItemSelectionModel.Deselect)
        assert_called([], [4])

        store.model.setSortIndices([1, 0, 3, 4, 2])
        store.model.setSortIndices(None)
        store.select_rows({2, 3}, QItemSelectionModel.Deselect)
        assert_called([], [3])


class SharedSelectionModelTest(SelectionModelTest):
    def test_select(self):
        self.store.select = Mock()
        sel = self.itsel({1, 2})
        self.model1.select(sel, QItemSelectionModel.Deselect)
        self.store.select.assert_called_with(sel, QItemSelectionModel.Deselect)

    def test_selection_from_rows(self):
        sel = self.model1.selection_from_rows({1, 2})
        self.assertEqual(len(sel), 2)
        ind1 = sel[0]
        self.assertIs(ind1.model(), self.model1.model())
        self.assertEqual(ind1.left(), 0)
        self.assertEqual(ind1.right(), self.model1.model().columnCount() - 1)
        self.assertEqual(ind1.top(), 1)
        self.assertEqual(ind1.bottom(), 1)

    def test_emit_selection_rows_changed(self):
        m1 = Mock()
        m2 = Mock()
        self.model1.selectionChanged.connect(m1)
        self.model2.selectionChanged.connect(m2)
        self.model1.model().setSortIndices(np.array([1, 0, 2, 4, 3]))
        self.model1.select(self.itsel({1, 3}), QItemSelectionModel.Select)
        self.assertEqual(self.store.rows, {0, 4})

        for model, m in zip((self.model1, self.model2), (m1, m2)):
            sel = m.call_args[0][0]
            self.assertEqual(len(sel), 2)
            for ind, row in zip(sel, (1, 3)):
                self.assertIs(ind.model(), model.model())
                self.assertEqual(ind.left(), 0)
                self.assertEqual(ind.right(), model.model().columnCount() - 1)
                self.assertEqual(ind.top(), row)
                self.assertEqual(ind.bottom(), row)

    def test_methods(self):
        def rowcol(sel):
            return {(index.row(), index.column()) for index in sel}

        self.model1.model().setSortIndices(np.array([1, 0, 2, 4, 3]))
        self.model2.model().setSortIndices(np.array([1, 0, 2, 4, 3]))

        self.assertFalse(self.model1.hasSelection())
        self.assertFalse(self.model1.isColumnSelected(1))
        self.assertFalse(self.model1.isRowSelected(1))
        self.assertEqual(self.model1.selectedColumns(1), [])
        self.assertEqual(self.model1.selectedRows(1), [])
        self.assertEqual(self.model1.selectedIndexes(), [])

        self.model1.select(self.itsel({1, 3}), QItemSelectionModel.Select)

        self.assertEqual(rowcol(self.model1.selection().indexes()),
                         {(1, 0), (1, 1), (3, 0), (3, 1)})
        self.assertEqual(rowcol(self.model2.selection().indexes()),
                         {(1, 0), (1, 1), (1, 2), (3, 0), (3, 1), (3, 2)})

        self.assertTrue(self.model1.hasSelection())
        self.assertFalse(self.model1.isColumnSelected(1))
        self.assertTrue(self.model1.isRowSelected(1))
        self.assertFalse(self.model1.isRowSelected(2))
        self.assertEqual(self.model1.selectedColumns(1), [])
        self.assertEqual(rowcol(self.model1.selectedRows(1)), {(1, 1), (3, 1)})
        self.assertEqual(rowcol(self.model1.selectedIndexes()),
                         {(1, 0), (1, 1), (3, 0), (3, 1)})

        self.model1.select(self.itsel({0, 1, 2, 4}), QItemSelectionModel.Select)
        self.assertTrue(self.model1.isColumnSelected(1))
        self.assertEqual(rowcol(self.model1.selectedColumns(1)),
                         {(1, 0), (1, 1)})

        self.model1.clearSelection()
        self.assertFalse(self.model1.hasSelection())
        self.assertFalse(self.model1.isColumnSelected(1))
        self.assertEqual(self.model1.selectedColumns(1), [])
        self.assertEqual(self.model1.selectedRows(1), [])
        self.assertEqual(self.model1.selectedIndexes(), [])

        self.model1.select(self.itsel({0, 1, 2, 3, 4}),
                           QItemSelectionModel.Select)
        self.model1.reset()
        self.assertFalse(self.model1.hasSelection())
        self.assertFalse(self.model1.isColumnSelected(1))
        self.assertEqual(self.model1.selectedColumns(1), [])
        self.assertEqual(self.model1.selectedRows(1), [])
        self.assertEqual(self.model1.selectedIndexes(), [])


class PredictionsModelTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.values = np.array([[0, 1, 1, 2, 0], [0, 0, 0, 1, 0]], dtype=float)
        cls.actual = np.array([0, 1, 2, 1, 0], dtype=float)
        cls.probs = [np.array([[80, 10, 10],
                               [30, 70, 0],
                               [15, 80, 5],
                               [0, 10, 90],
                               [55, 40, 5]]) / 100,
                     np.array([[80, 0, 20],
                               [90, 5, 5],
                               [70, 10, 20],
                               [10, 60, 30],
                               [50, 25, 25]]) / 100]
        cls.no_probs = [np.zeros((5, 0)), np.zeros((5, 0))]

    def test_model_classification(self):
        model = PredictionsModel(self.values, self.probs, self.actual)
        self.assertEqual(model.rowCount(), 5)
        self.assertEqual(model.columnCount(), 4)

        val, prob = model.data(model.index(0, 2))
        self.assertEqual(val, 0)
        np.testing.assert_equal(prob, [0.8, 0, 0.2])

        val, prob = model.data(model.index(3, 2))
        self.assertEqual(val, 1)
        np.testing.assert_equal(prob, [0.1, 0.6, 0.3])

    def test_model_classification_errors(self):
        model = PredictionsModel(self.values, self.probs, self.actual)

        np.testing.assert_almost_equal(
            [model.data(model.index(row, 1)) for row in range(5)],
            1 - np.array([80, 70, 5, 10, 55]) / 100)

        np.testing.assert_almost_equal(
            [model.data(model.index(row, 3)) for row in range(5)],
            1 - np.array([80, 5, 20, 60, 50]) / 100)

    def test_model_regression(self):
        model = PredictionsModel(self.values, self.no_probs, self.actual)
        self.assertEqual(model.rowCount(), 5)
        self.assertEqual(model.columnCount(), 4)

        val, prob = model.data(model.index(0, 2))
        self.assertEqual(val, 0)
        np.testing.assert_equal(prob, [])

        val, prob = model.data(model.index(3, 2))
        self.assertEqual(val, 1)
        np.testing.assert_equal(prob, [])

    def test_model_regression_errors(self):
        actual = np.array([40, 0, 12, 0, -45])
        model = PredictionsModel(values=np.array([[0] * 5,
                                                  [30, 0, 12, -5, -40]]),
                                 probs=self.no_probs,
                                 actual=actual,
                                 reg_error_type=NO_ERR)

        self.assertIsNone(model.data(model.index(0, 1)))

        model.setRegErrorType(DIFF_ERROR)
        diff_error = np.array([-10, 0, 0, -5, 5])
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 3)) for row in range(5)],
            diff_error)
        np.testing.assert_almost_equal(model.errorColumn(1), diff_error)

        model.setRegErrorType(ABSDIFF_ERROR)
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 3)) for row in range(5)],
            np.abs(diff_error))
        np.testing.assert_almost_equal(model.errorColumn(1), np.abs(diff_error))

        model.setRegErrorType(REL_ERROR)
        rel_error = [-10 / 40, 0, 0, -np.inf, 5 / 45]
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 3)) for row in range(5)], rel_error)
        np.testing.assert_almost_equal(model.errorColumn(1), rel_error)

        model.setRegErrorType(ABSREL_ERROR)
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 3)) for row in range(5)], np.abs(rel_error))
        np.testing.assert_almost_equal(model.errorColumn(1), np.abs(rel_error))

        model.setRegErrorType(DIFF_ERROR)
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 1)) for row in range(5)], -actual)
        np.testing.assert_almost_equal(model.errorColumn(0), -actual)

    def test_model_actual(self):
        model = PredictionsModel(self.values, self.no_probs, self.actual)
        self.assertEqual(model.data(model.index(2, 0), Qt.UserRole),
                         self.actual[2])

    def test_model_no_actual(self):
        model = PredictionsModel(self.values, self.no_probs, None)
        self.assertTrue(np.isnan(model.data(model.index(2, 0), Qt.UserRole)),
                         self.actual[2])

    def test_model_header(self):
        model = PredictionsModel(self.values, self.probs, self.actual)
        self.assertIsNone(model.headerData(0, Qt.Horizontal))
        self.assertEqual(model.headerData(3, Qt.Vertical), "4")

        model = PredictionsModel(self.values, self.probs, self.actual, ["a", "b"])
        self.assertEqual(model.headerData(0, Qt.Horizontal), "a")
        self.assertEqual(model.headerData(1, Qt.Horizontal), "error")
        self.assertEqual(model.headerData(2, Qt.Horizontal), "b")
        self.assertEqual(model.headerData(3, Qt.Horizontal), "error")
        self.assertIsNone(model.headerData(4, Qt.Horizontal))
        self.assertEqual(model.headerData(4, Qt.Vertical), "5")

        model = PredictionsModel(self.values, self.probs, self.actual, ["a"])
        self.assertEqual(model.headerData(0, Qt.Horizontal), "a")
        self.assertEqual(model.headerData(1, Qt.Horizontal), "error")
        self.assertIsNone(model.headerData(2, Qt.Horizontal))
        self.assertEqual(model.headerData(3, Qt.Vertical), "4")

    def test_model_empty(self):
        model = PredictionsModel()
        self.assertEqual(model.rowCount(), 0)
        self.assertEqual(model.columnCount(), 0)
        self.assertIsNone(model.headerData(1, Qt.Horizontal))

    def test_sorting_classification(self):
        model = PredictionsModel(self.values, self.probs, self.actual)

        val, prob = model.data(model.index(0, 2))
        self.assertEqual(val, 0)
        np.testing.assert_equal(prob, [0.8, 0, 0.2])

        val, prob = model.data(model.index(3, 2))
        self.assertEqual(val, 1)
        np.testing.assert_equal(prob, [0.1, 0.6, 0.3])

        model.setProbInd([[2], [2]])
        model.sort(0, Qt.DescendingOrder)
        val, prob = model.data(model.index(0, 0))
        self.assertEqual(val, 2)
        np.testing.assert_equal(prob, [0, 0.1, 0.9])
        val, prob = model.data(model.index(0, 2))
        self.assertEqual(val, 1)
        np.testing.assert_equal(prob, [0.1, 0.6, 0.3])

        model.setProbInd([[2], [2]])
        model.sort(2, Qt.AscendingOrder)
        val, prob = model.data(model.index(0, 2))
        self.assertEqual(val, 0)
        np.testing.assert_equal(prob, [0.9, 0.05, 0.05])
        val, prob = model.data(model.index(0, 0))
        self.assertEqual(val, 1)
        np.testing.assert_equal(prob, [0.3, 0.7, 0])

        model.setProbInd([[1, 0], [1, 0]])
        model.sort(0, Qt.AscendingOrder)
        np.testing.assert_equal(model.data(model.index(0, 0))[1], [0, .1, .9])
        np.testing.assert_equal(model.data(model.index(1, 0))[1], [0.8, .1, .1])

        model.setProbInd([[1, 2], [1, 2]])
        model.sort(0, Qt.AscendingOrder)
        np.testing.assert_equal(model.data(model.index(0, 0))[1], [0.8, .1, .1])
        np.testing.assert_equal(model.data(model.index(1, 0))[1], [0, .1, .9])

        model.setProbInd([[], []])
        model.sort(0, Qt.AscendingOrder)
        self.assertEqual([model.data(model.index(i, 0))[0]
                          for i in range(model.rowCount())], [0, 0, 1, 1, 2])

        model.setProbInd([[], []])
        model.sort(0, Qt.DescendingOrder)
        self.assertEqual([model.data(model.index(i, 0))[0]
                          for i in range(model.rowCount())], [2, 1, 1, 0, 0])

    def test_sorting_classification_error(self):
        model = PredictionsModel(self.values, self.probs, self.actual)

        np.testing.assert_almost_equal(
            [model.data(model.index(row, 1)) for row in range(5)],
            1 - np.array([80, 70, 5, 10, 55]) / 100)

        np.testing.assert_almost_equal(
            [model.data(model.index(row, 3)) for row in range(5)],
            1 - np.array([80, 5, 20, 60, 50]) / 100)

        model.sort(1, Qt.AscendingOrder)
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 1)) for row in range(5)],
            1 - np.array(sorted([80, 70, 5, 10, 55], reverse=True)) / 100)

        model.sort(3, Qt.DescendingOrder)
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 3)) for row in range(5)],
            1 - np.array(sorted([80, 5, 20, 60, 50])) / 100)

    def test_sorting_classification_different(self):
        model = PredictionsModel(self.values, self.probs, self.actual)

        model.setProbInd([[2], [0]])
        model.sort(0, Qt.DescendingOrder)
        val, prob = model.data(model.index(0, 0))
        self.assertEqual(val, 2)
        np.testing.assert_equal(prob, [0, 0.1, 0.9])
        val, prob = model.data(model.index(0, 2))
        self.assertEqual(val, 1)
        np.testing.assert_equal(prob, [0.1, 0.6, 0.3])
        model.sort(2, Qt.DescendingOrder)
        val, prob = model.data(model.index(0, 0))
        self.assertEqual(val, 1)
        np.testing.assert_equal(prob, [0.3, 0.7, 0])
        val, prob = model.data(model.index(0, 2))
        self.assertEqual(val, 0)
        np.testing.assert_equal(prob, [0.9, 0.05, 0.05])

    def test_sorting_regression(self):
        model = PredictionsModel(self.values, self.no_probs, self.actual)

        self.assertEqual(model.data(model.index(0, 2))[0], 0)
        self.assertEqual(model.data(model.index(3, 2))[0], 1)

        model.setProbInd([2])
        model.sort(0, Qt.AscendingOrder)
        self.assertEqual([model.data(model.index(i, 0))[0]
                          for i in range(model.rowCount())], [0, 0, 1, 1, 2])

        model.setProbInd([])
        model.sort(0, Qt.DescendingOrder)
        self.assertEqual([model.data(model.index(i, 0))[0]
                          for i in range(model.rowCount())], [2, 1, 1, 0, 0])

        model.setProbInd(None)
        model.sort(0, Qt.AscendingOrder)
        self.assertEqual([model.data(model.index(i, 0))[0]
                          for i in range(model.rowCount())], [0, 0, 1, 1, 2])

    def test_sorting_regression_error(self):
        actual = np.array([40, 0, 12, 0, -45])
        model = PredictionsModel(values=np.array([[30, 0, 12, -5, -40]]),
                                 probs=self.no_probs[:1],
                                 actual=actual,
                                 reg_error_type=NO_ERR)

        model.setRegErrorType(DIFF_ERROR)
        model.sort(1, Qt.AscendingOrder)
        diff_error = [-10, 0, 0, -5, 5]
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 1)) for row in range(5)],
            sorted(diff_error))

        model.setRegErrorType(ABSDIFF_ERROR)
        model.sort(1, Qt.AscendingOrder)
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 1)) for row in range(5)],
            sorted(np.abs(diff_error)))

        model.setRegErrorType(REL_ERROR)
        rel_error = [-10 / 40, 0, 0, -np.inf, 5 / 45]
        model.sort(1, Qt.AscendingOrder)
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 1)) for row in range(5)],
            sorted(rel_error))

        model.setRegErrorType(ABSREL_ERROR)
        model.sort(1, Qt.AscendingOrder)
        np.testing.assert_almost_equal(
            [model.data(model.index(row, 1)) for row in range(5)],
            sorted(np.abs(rel_error)))


class TestPredictionsItemDelegate(GuiTest):
    def test_displayText(self):
        delegate = PredictionsItemDelegate()
        delegate.fmt = "{value:.3f}"
        self.assertEqual(delegate.displayText((0.12345, [1, 2, 3]), Mock()), "0.123")
        delegate.fmt = "{value:.1f}"
        self.assertEqual(delegate.displayText((0.12345, [1, 2, 3]), Mock()), "0.1")
        delegate.fmt = "{value:.1f} - {dist[2]}"
        self.assertEqual(delegate.displayText((0.12345, [1, 2, 3]), Mock()), "0.1 - 3")

        self.assertEqual(delegate.displayText(None, Mock()), "")


class TestClassificationItemDelegate(GuiTest):
    @patch.object(QToolTip, "showText")
    def test_format(self, showText):
        delegate = ClassificationItemDelegate(["foo", "bar", "baz"],
                                              [(1, 2, 3), (4, 5, 6), (7, 8, 9)],
                                              (1, None, 0), ("foo", "baz")
                                              )
        delegate.helpEvent(Mock(), Mock(), Mock(), Mock())
        self.assertEqual(showText.call_args[0][1], "p(foo, baz)")
        self.assertEqual(delegate.displayText((["baz", (0.4, 0.6)]), Mock()),
                         "0.60 : - : 0.40 → baz")
        showText.reset_mock()

        delegate = ClassificationItemDelegate(["foo", "bar", "baz"],
                                              [(1, 2, 3), (4, 5, 6), (7, 8, 9)],
                                              )
        delegate.helpEvent(Mock(), Mock(), Mock(), Mock())
        self.assertEqual(showText.call_args[0][1], "")
        self.assertEqual(delegate.displayText((["baz", (0.4, 0.6)]), Mock()),
                         "baz")

    def test_drawbar(self):
        delegate = ClassificationItemDelegate(
            ["foo", "bar", "baz", "bax"],
            [(1, 2, 3), (4, 5, 6), (7, 8, 9), (10, 11, 12)],
            (1, None, 0, 2), ("baz", "foo", "bax"))
        painter = Mock()
        rr = painter.drawRoundedRect
        index = Mock()
        index.data = lambda *_: 2
        rect = QRect(0, 0, 256, 16)

        delegate.cachedData = lambda *_: None
        delegate.drawBar(painter, Mock(), index, rect)
        rr.assert_not_called()

        delegate.cachedData = lambda *_: (1, (0.25, 0, 0.75, 0))
        delegate.drawBar(painter, Mock(), index, rect)
        self.assertEqual(rr.call_count, 2)
        rect = rr.call_args_list[0][0][0]
        self.assertEqual(rect.width(), 64)
        self.assertEqual(rect.height(), 8)
        rect = rr.call_args_list[1][0][0]
        self.assertEqual(rect.width(), 192)
        self.assertEqual(rect.height(), 16)


class TestNoopItemDelegate(GuiTest):
    def test_donothing(self):
        delegate = NoopItemDelegate()
        delegate.paint(Mock(), Mock(), Mock(), Mock())
        delegate.sizeHint()


class TestRegressionItemDelegate(GuiTest):
    def test_format(self):
        delegate = RegressionItemDelegate("%6.3f")
        self.assertEqual(delegate.displayText((5.13, None), Mock()), " 5.130")
        self.assertEqual(delegate.offset, 0)
        self.assertEqual(delegate.span, 1)

        delegate = RegressionItemDelegate("%6.3f", 2, 5)
        self.assertEqual(delegate.displayText((5.13, None), Mock()), " 5.130")
        self.assertEqual(delegate.offset, 2)
        self.assertEqual(delegate.span, 3)

        delegate = RegressionItemDelegate(None, 2, 5)
        self.assertEqual(delegate.displayText((5.1, None), Mock()), "5.10")
        self.assertEqual(delegate.offset, 2)
        self.assertEqual(delegate.span, 3)

    def test_drawBar(self):
        delegate = RegressionItemDelegate("%6.3f", 2, 10)
        painter = Mock()
        dr = painter.drawRect
        el = painter.drawEllipse
        index = Mock()
        rect = QRect(0, 0, 256, 16)

        ### Actual is missing
        index.data = lambda *_: np.nan

        # Prediction is missing
        delegate.cachedData = lambda *_: None
        delegate.drawBar(painter, Mock(), index, rect)

        dr.assert_not_called()
        el.assert_not_called()

        # Prediction is known
        delegate.cachedData = lambda *_: (8.0, None)
        delegate.drawBar(painter, Mock(), index, rect)

        dr.assert_called_once()
        rrect = dr.call_args[0][0]
        self.assertEqual(rrect.width(), 192)
        el.assert_not_called()
        dr.reset_mock()

        ### Actual is known
        index.data = lambda *_: 8.0

        # Prediction is correct
        delegate.cachedData = lambda *_: (8.0, None)
        delegate.drawBar(painter, Mock(), index, rect)

        dr.assert_called_once()
        rrect = dr.call_args[0][0]
        self.assertEqual(rrect.width(), 192)
        el.assert_called_once()
        center = el.call_args[0][0]
        self.assertEqual(center.x(), 192)
        dr.reset_mock()
        el.reset_mock()

        # Prediction is below
        delegate.cachedData = lambda *_: (6.0, None)
        delegate.drawBar(painter, Mock(), index, rect)

        dr.assert_called_once()
        rrect = dr.call_args[0][0]
        self.assertEqual(rrect.width(), 128)
        el.assert_called_once()
        center = el.call_args[0][0]
        self.assertEqual(center.x(), 192)
        dr.reset_mock()
        el.reset_mock()

        # Prediction is above
        delegate.cachedData = lambda *_: (9.0, None)
        delegate.drawBar(painter, Mock(), index, rect)

        dr.assert_called_once()
        rrect = dr.call_args[0][0]
        self.assertEqual(rrect.width(), 224)
        el.assert_called_once()
        center = el.call_args[0][0]
        self.assertEqual(center.x(), 192)
        dr.reset_mock()
        el.reset_mock()


class TestClassificationErrorDelegate(GuiTest):
    def test_displayText(self):
        delegate = ClassificationErrorDelegate()
        self.assertEqual(delegate.displayText(0.12345, Mock()), "0.123")
        self.assertEqual(delegate.displayText(np.nan, Mock()), "?")

    def test_drawBar(self):
        delegate = ClassificationErrorDelegate()
        painter = Mock()
        dr = painter.drawRect
        index = Mock()
        rect = QRect(0, 0, 256, 16)

        delegate.cachedData = lambda *_: np.nan
        delegate.drawBar(painter, Mock(), index, rect)
        dr.assert_not_called()

        delegate.cachedData = lambda *_: None
        delegate.drawBar(painter, Mock(), index, rect)
        dr.assert_not_called()

        delegate.cachedData = lambda *_: 1 / 4
        delegate.drawBar(painter, Mock(), index, rect)
        dr.assert_called_once()
        r = dr.call_args[0][0]
        self.assertEqual(r.x(), 0)
        self.assertEqual(r.y(), 0)
        self.assertEqual(r.width(), 64)
        self.assertEqual(r.height(), 16)


class TestRegressionErrorDelegate(GuiTest):
    def test_displayText(self):
        delegate = RegressionErrorDelegate("", True, 4)
        self.assertEqual(delegate.displayText(0.1234567, Mock()), "")

        delegate = RegressionErrorDelegate("%.5f", True, 4)
        self.assertEqual(delegate.displayText(0.1234567, Mock()), "0.12346")
        self.assertEqual(delegate.displayText(np.nan, Mock()), "?")
        self.assertEqual(delegate.displayText(np.inf, Mock()), "∞")
        self.assertEqual(delegate.displayText(-np.inf, Mock()), "-∞")

    def test_drawBar(self):
        painter = Mock()
        dr = painter.drawRect
        index = Mock()
        rect = QRect(0, 0, 256, 16)

        delegate = RegressionErrorDelegate("%.5f", True, 0)
        delegate.drawBar(painter, Mock(), index, rect)
        dr.assert_not_called()

        delegate = RegressionErrorDelegate("%.5f", True, 12)

        delegate.cachedData = lambda *_: np.nan
        delegate.drawBar(painter, Mock(), index, rect)
        dr.assert_not_called()

        delegate.cachedData = lambda *_: None
        delegate.drawBar(painter, Mock(), index, rect)
        dr.assert_not_called()

        delegate.cachedData = lambda *_: 3
        delegate.drawBar(painter, Mock(), index, rect)
        r = dr.call_args[0][0]
        self.assertEqual(r.x(), 128)
        self.assertEqual(r.y(), 0)
        self.assertEqual(r.width(), 32)
        self.assertEqual(r.height(), 16)

        delegate.cachedData = lambda *_: -3
        delegate.drawBar(painter, Mock(), index, rect)
        r = dr.call_args[0][0]
        self.assertEqual(r.x(), 128)
        self.assertEqual(r.y(), 0)
        self.assertEqual(r.width(), -32)
        self.assertEqual(r.height(), 16)

        delegate = RegressionErrorDelegate("%.5f", False, 12)
        delegate.cachedData = lambda *_: 3
        delegate.drawBar(painter, Mock(), index, rect)
        r = dr.call_args[0][0]
        self.assertEqual(r.x(), 0)
        self.assertEqual(r.y(), 0)
        self.assertEqual(r.width(), 64)
        self.assertEqual(r.height(), 16)


if __name__ == "__main__":
    unittest.main()



# pylint: disable=protected-access

import unittest
from functools import partial
from unittest.mock import patch

import numpy as np
from AnyQt.QtCore import Qt

from orangewidget.settings import Context

from Orange.misc import DistMatrix
from Orange.data import Table, Domain, ContinuousVariable, StringVariable
from Orange.distance import Euclidean
from Orange.widgets.tests.base import WidgetTest
from Orange.widgets.unsupervised.owdistancematrix import OWDistanceMatrix


class TestOWDistanceMatrix(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWDistanceMatrix)
        self.iris = Table("iris")[:5]
        self.distances = Euclidean(self.iris)

    def test_set_distances(self):
        assert isinstance(self.widget, OWDistanceMatrix)

        # Distances with row data
        self.widget.set_distances(self.distances)
        self.assertIn(self.iris.domain[0], self.widget.annot_combo.model())
        self.widget.send_report()

        # Distances without row data
        self.distances.row_items = None
        self.widget.set_distances(self.distances)
        self.assertNotIn(self.iris.domain[0], self.widget.annot_combo.model())
        self.widget.send_report()

        # Non-square distances, no labels
        distances = DistMatrix(np.array([[1, 2, 3], [4, 5, 6]]))
        self.widget.set_distances(distances)
        self.assertEqual(self.widget.annot_combo.model().rowCount(), 2)
        self.widget.send_report()

        # Non-square distances, row labels
        distances = DistMatrix(np.array([[1, 2, 3], [4, 5, 6]]))
        distances.row_items = list("ab")
        self.widget.set_distances(distances)
        self.assertEqual(self.widget.annot_combo.model().rowCount(), 3)
        self.widget.send_report()

        # Non-square distances, column labels
        distances = DistMatrix(np.array([[1, 2, 3], [4, 5, 6]]))
        distances.col_items = list("def")
        self.widget.set_distances(distances)
        self.assertEqual(self.widget.annot_combo.model().rowCount(), 3)
        self.widget.send_report()

        # Non-square distances, both labels
        distances = DistMatrix(np.array([[1, 2, 3], [4, 5, 6]]))
        distances.row_items = list("ab")
        distances.col_items = list("def")
        self.widget.set_distances(distances)
        self.assertEqual(self.widget.annot_combo.model().rowCount(), 3)
        self.widget.send_report()

        # Non-square distances, no labels
        distances = DistMatrix(np.array([[1, 2, 3], [4, 5, 6]]))
        self.widget.set_distances(distances)
        self.assertEqual(self.widget.annot_combo.model().rowCount(), 2)
        self.widget.send_report()

    def test_context_attribute(self):
        distances = Euclidean(self.iris, axis=0)
        annotations = ["None", "Enumerate"]
        self.widget.set_distances(distances)
        self.widget.openContext(distances, annotations)

    def test_unconditional_commit_on_new_signal(self):
        with patch.object(self.widget.commit, 'now') as commit:
            self.widget.auto_commit = False
            commit.reset_mock()
            self.send_signal(self.widget.Inputs.distances, self.distances)
            commit.assert_called()

    def test_labels(self):
        x, y = (ContinuousVariable(c) for c in "xy")
        s = StringVariable("s")
        grades = Table.from_list(
            Domain([x, y], [], [s]),
            [[91.0, 89.0, "Bill"],
             [51.0, 100.0, "Cynthia"],
             [9.0, 61.0, "Demi"],
             [49.0, 92.0, "Fred"],
             [91.0, 49.0, "George"]
             ]
        )

        header = self.widget.tablemodel.headerData

        distances = Euclidean(grades)
        self.widget.set_distances(distances)
        ac = self.widget.annot_combo
        idx = ac.model().indexOf(grades.domain.metas[0])
        ac.setCurrentIndex(idx)
        ac.activated.emit(idx)
        self.assertEqual(header(2, Qt.Horizontal, Qt.DisplayRole), "Demi")
        self.assertIsNone(header(2, Qt.Horizontal, Qt.BackgroundRole))
        self.assertEqual(header(2, Qt.Vertical, Qt.DisplayRole), "Demi")
        self.assertIsNone(header(2, Qt.Vertical, Qt.BackgroundRole))

        idx = ac.model().indexOf(grades.domain.attributes[0])
        ac.setCurrentIndex(idx)
        ac.activated.emit(idx)
        self.assertIn("9", header(2, Qt.Horizontal, Qt.DisplayRole))
        self.assertIsNotNone(header(2, Qt.Horizontal, Qt.BackgroundRole))
        self.assertIn("9", header(2, Qt.Vertical, Qt.DisplayRole))
        self.assertIsNotNone(header(2, Qt.Vertical, Qt.BackgroundRole))

    def test_num_meta_labels_w_nan(self):
        x, y = (ContinuousVariable(c) for c in "xy")
        s = StringVariable("s")
        data = Table.from_list(
            Domain([x], [], [y, s]),
            [[0, 1, "a"],
             [1, np.nan, "b"]]
        )
        distances = Euclidean(data)
        self.widget.set_distances(distances)
        ac = self.widget.annot_combo
        idx = ac.model().indexOf(y)
        ac.setCurrentIndex(idx)
        ac.activated.emit(idx)

        header = self.widget.tablemodel.headerData
        self.assertEqual(header(0, Qt.Horizontal, Qt.DisplayRole), "1")
        self.assertEqual(header(1, Qt.Horizontal, Qt.DisplayRole), "?")
        self.assertIsNotNone(header(1, Qt.Horizontal, Qt.BackgroundRole))
        self.assertEqual(header(0, Qt.Vertical, Qt.DisplayRole), "1")
        self.assertEqual(header(1, Qt.Vertical, Qt.DisplayRole), "?")
        self.assertIsNotNone(header(1, Qt.Vertical, Qt.BackgroundRole))

    def test_choose_label(self):
        self.assertIs(OWDistanceMatrix._choose_label(self.iris),
                      self.iris.domain.class_var)

        domain = Domain([ContinuousVariable(x) for x in "xyz"],
                        ContinuousVariable("t"),
                        [ContinuousVariable("m")] +
                        [StringVariable(c) for c in "abc"]
                        )
        data = Table.from_numpy(
            domain,
            np.zeros((4, 3), dtype=float),
            np.arange(4, dtype=float),
            np.array([[0, "a", "a", "a"],
                      [1, "b", "b", "b"],
                      [2, "a", "c", "b"],
                      [0, "b", "a", "a"]])
        )
        self.assertIs(OWDistanceMatrix._choose_label(data),
                      domain.metas[2])
        domain2 = Domain(domain.attributes, domain.class_var, domain.metas[:-2])
        self.assertIs(OWDistanceMatrix._choose_label(data.transform(domain2)),
                      domain.metas[1])

    def test_non_square_labels(self):
        widget = self.widget
        ac = self.widget.annot_combo

        dist = DistMatrix([[1, 2, 3], [4, 5, 6]])
        dist.row_items = DistMatrix._labels_to_tables(["aa", "bb"])
        dist.col_items = DistMatrix._labels_to_tables(["cc", "dd", "ee"])
        self.send_signal(widget.Inputs.distances, dist)
        self.assertEqual(ac.model().rowCount(), 3)

        header = partial(widget.tablemodel.headerData, role=Qt.DisplayRole)
        ac.setCurrentIndex(0)
        ac.activated.emit(0)
        self.assertIsNone(header(1, Qt.Horizontal))
        self.assertIsNone(header(1, Qt.Vertical))

        ac.setCurrentIndex(1)
        ac.activated.emit(1)
        self.assertEqual(header(1, Qt.Horizontal), "2")
        self.assertEqual(header(1, Qt.Vertical), "2")

        ac.setCurrentIndex(2)
        ac.activated.emit(2)
        self.assertEqual(header(1, Qt.Horizontal), "dd")
        self.assertEqual(header(1, Qt.Vertical), "bb")

    @WidgetTest.skipNonEnglish
    def test_migrate_settings_v1_and_use_them(self):
        ind = [1, 2, 5, 6, 7, 8]
        context = Context(
            values={'__version__': 1},
            dim=10,
            annotations=['None', 'Enumerate',
                         'sepal length', 'sepal width', 'petal length',
                         'petal width', 'iris'],
            annotation='petal length',
            selection=ind)
        widget = self.create_widget(
            OWDistanceMatrix, stored_settings={"__version__": 1, "context_settings": [context]})
        iris = Table("iris")[:10]
        distances = Euclidean(iris)
        self.send_signal(widget.Inputs.distances, distances)
        self.assertEqual(widget.annotation_idx, 4)
        self.assertEqual(widget.tableview.selectionModel().selectedItems(), ind)
        outm = self.get_output(widget.Outputs.distances)
        np.testing.assert_equal(outm, distances.submatrix(ind, ind))

    def test_square_settings(self):
        widget = self.widget
        self.send_signal(widget.Inputs.distances, self.distances)
        widget._set_selection([0, 3, 4])
        widget.annotation_idx = 3

        self.send_signal(widget.Inputs.distances, None)
        self.assertEqual(widget._get_selection(), ([], True))
        self.assertEqual(widget.annotation_idx, 0)

        self.send_signal(widget.Inputs.distances, self.distances)
        self.assertEqual(widget._get_selection(), ([0, 3, 4], True))
        self.assertEqual(widget.annotation_idx, 3)

        matrix = DistMatrix(np.array([[1, 2, 3], [4, 5, 6]]))
        matrix.row_items = list("ab")

        self.send_signal(widget.Inputs.distances,matrix)
        self.assertEqual(widget._get_selection(), (([], []), False))
        self.assertEqual(widget.annotation_idx, 2)

        widget._set_selection(([0], [0, 2]))
        widget.annotation_idx = 0

        self.send_signal(widget.Inputs.distances, self.distances)
        self.assertEqual(widget._get_selection(), ([0, 3, 4], True))
        self.assertEqual(widget.annotation_idx, 3)

        self.send_signal(widget.Inputs.distances, matrix)
        self.assertEqual(widget._get_selection(), (([0], [0, 2]), False))
        self.assertEqual(widget.annotation_idx, 0)


if __name__ == "__main__":
    unittest.main()
